"""
Algorithm synthesis chain for extracting and formalizing algorithms.
"""

import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from ..core.retrieval import RetrievalService, RetrievalResult
from ..core.models import LLMClient, GenerationConfig
from ..core.prompts import get_prompt
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class Algorithm:
    """Represents an extracted algorithm."""
    name: str
    purpose: str = ""
    input_spec: str = ""
    output_spec: str = ""
    pseudocode: str = ""
    time_complexity: str = ""
    space_complexity: str = ""
    key_insight: str = ""
    edge_cases: List[str] = field(default_factory=list)
    source_sections: List[str] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "purpose": self.purpose,
            "input": self.input_spec,
            "output": self.output_spec,
            "pseudocode": self.pseudocode,
            "time_complexity": self.time_complexity,
            "space_complexity": self.space_complexity,
            "key_insight": self.key_insight,
            "edge_cases": self.edge_cases,
            "source_sections": self.source_sections
        }
    
    @property
    def complexity_str(self) -> str:
        """Get formatted complexity string."""
        parts = []
        if self.time_complexity:
            parts.append(f"Time: {self.time_complexity}")
        if self.space_complexity:
            parts.append(f"Space: {self.space_complexity}")
        return ", ".join(parts) if parts else ""


class AlgorithmSynthesizer:
    """
    Chain for extracting and synthesizing algorithms.
    
    Generates:
    - Structured pseudocode
    - Complexity analysis
    - Key insights
    - Edge cases
    """
    
    def __init__(
        self,
        retrieval_service: RetrievalService,
        llm_client: LLMClient,
        include_pseudocode: bool = True,
        include_complexity: bool = True,
        include_examples: bool = True
    ):
        """
        Initialize algorithm synthesizer.
        
        Args:
            retrieval_service: Service for retrieving context
            llm_client: LLM for generation
            include_pseudocode: Generate pseudocode
            include_complexity: Include complexity analysis
            include_examples: Include examples
        """
        self.retrieval_service = retrieval_service
        self.llm_client = llm_client
        self.include_pseudocode = include_pseudocode
        self.include_complexity = include_complexity
        self.include_examples = include_examples
    
    @trace_function(name="extract_algorithms")
    def extract(
        self,
        queries: Optional[List[str]] = None,
        config: Optional[GenerationConfig] = None
    ) -> List[Algorithm]:
        """
        Extract algorithms from the indexed document.
        
        Args:
            queries: Optional specific queries
            config: Generation configuration
            
        Returns:
            List of extracted algorithms
        """
        config = config or GenerationConfig()
        
        # Default queries for algorithm extraction
        if queries is None:
            queries = [
                "algorithm procedure steps",
                "pseudocode implementation",
                "training algorithm optimization",
                "method procedure approach",
            ]
        
        all_algorithms = []
        seen_names = set()
        
        for query in queries:
            # Retrieve code/algorithm content
            results = self.retrieval_service.retrieve_code(query, top_k=5)
            
            if not results:
                results = self.retrieval_service.retrieve(query, top_k=5)
            
            if not results:
                continue
            
            # Extract algorithms from context
            algorithms = self._extract_from_context(results, config)
            
            # Deduplicate
            for algo in algorithms:
                name_key = algo.name.lower().replace(" ", "_")
                if name_key not in seen_names and algo.pseudocode:
                    all_algorithms.append(algo)
                    seen_names.add(name_key)
        
        logger.info(f"Extracted {len(all_algorithms)} algorithms")
        return all_algorithms
    
    def _extract_from_context(
        self,
        results: List[RetrievalResult],
        config: GenerationConfig
    ) -> List[Algorithm]:
        """Extract algorithms from retrieval results."""
        context = "\n\n".join(r.content for r in results)
        
        # Track source sections
        source_sections = list(set(
            section
            for r in results
            for section in r.section_hierarchy
            if section
        ))
        
        prompt = get_prompt("algorithm_extraction")(context=context)
        result = self.llm_client.generate(prompt, config)
        
        # Parse response
        algorithms = self._parse_response(result.text, source_sections)
        
        return algorithms
    
    def _parse_response(
        self,
        text: str,
        source_sections: List[str]
    ) -> List[Algorithm]:
        """Parse algorithm extraction response."""
        algorithms = []
        current: Dict[str, Any] = {}
        in_pseudocode = False
        pseudocode_lines = []
        
        for line in text.split("\n"):
            line_stripped = line.strip()
            
            if line_stripped.startswith("ALGORITHM:"):
                # Save previous algorithm
                if current:
                    if pseudocode_lines:
                        current["pseudocode"] = "\n".join(pseudocode_lines)
                    algorithms.append(self._create_algorithm(current, source_sections))
                
                current = {"name": line_stripped[10:].strip()}
                pseudocode_lines = []
                in_pseudocode = False
                
            elif line_stripped.startswith("PURPOSE:"):
                current["purpose"] = line_stripped[8:].strip()
                
            elif line_stripped.startswith("INPUT:"):
                current["input"] = line_stripped[6:].strip()
                
            elif line_stripped.startswith("OUTPUT:"):
                current["output"] = line_stripped[7:].strip()
                
            elif line_stripped.startswith("PSEUDOCODE:"):
                in_pseudocode = True
                
            elif line_stripped == "```" or line_stripped == "```python":
                in_pseudocode = not in_pseudocode
                
            elif in_pseudocode:
                pseudocode_lines.append(line.rstrip())
                
            elif line_stripped.startswith("COMPLEXITY:"):
                # Parse complexity
                complexity_text = line_stripped[11:].strip()
                current["complexity"] = complexity_text
                in_pseudocode = False
                
            elif line_stripped.startswith("- Time:"):
                current["time_complexity"] = line_stripped[7:].strip()
                
            elif line_stripped.startswith("- Space:"):
                current["space_complexity"] = line_stripped[8:].strip()
                
            elif line_stripped.startswith("KEY INSIGHT:"):
                current["key_insight"] = line_stripped[12:].strip()
                in_pseudocode = False
                
            elif line_stripped.startswith("EDGE CASES:"):
                current["edge_cases"] = line_stripped[11:].strip()
                in_pseudocode = False
        
        # Don't forget the last algorithm
        if current:
            if pseudocode_lines:
                current["pseudocode"] = "\n".join(pseudocode_lines)
            algorithms.append(self._create_algorithm(current, source_sections))
        
        return algorithms
    
    def _create_algorithm(
        self,
        data: Dict[str, Any],
        source_sections: List[str]
    ) -> Algorithm:
        """Create an Algorithm from parsed data."""
        # Parse edge cases if string
        edge_cases = data.get("edge_cases", [])
        if isinstance(edge_cases, str):
            edge_cases = [e.strip() for e in edge_cases.split(",") if e.strip()]
        
        # Parse complexity if combined
        time_complexity = data.get("time_complexity", "")
        space_complexity = data.get("space_complexity", "")
        
        if "complexity" in data and not time_complexity:
            complexity = data["complexity"]
            if "Time:" in complexity:
                time_complexity = complexity.split("Time:")[1].split(",")[0].strip()
            if "Space:" in complexity:
                space_complexity = complexity.split("Space:")[1].strip()
        
        return Algorithm(
            name=data.get("name", ""),
            purpose=data.get("purpose", ""),
            input_spec=data.get("input", ""),
            output_spec=data.get("output", ""),
            pseudocode=data.get("pseudocode", ""),
            time_complexity=time_complexity,
            space_complexity=space_complexity,
            key_insight=data.get("key_insight", ""),
            edge_cases=edge_cases,
            source_sections=source_sections
        )
    
    @trace_function(name="synthesize_algorithm")
    def synthesize(
        self,
        description: str,
        algorithm_name: str,
        config: Optional[GenerationConfig] = None
    ) -> Algorithm:
        """
        Synthesize pseudocode from a description.
        
        Args:
            description: Natural language description
            algorithm_name: Name for the algorithm
            config: Generation configuration
            
        Returns:
            Synthesized Algorithm
        """
        config = config or GenerationConfig()
        
        prompt = get_prompt("algorithm_synthesis")(
            description=description,
            algorithm_name=algorithm_name
        )
        
        result = self.llm_client.generate(prompt, config)
        
        # Parse the synthesized algorithm
        algorithms = self._parse_response(result.text, [])
        
        if algorithms:
            return algorithms[0]
        
        # Fallback: create basic algorithm
        return Algorithm(
            name=algorithm_name,
            purpose=description,
            pseudocode=result.text
        )
    
    def format_for_latex(self, algorithm: Algorithm) -> str:
        """
        Format algorithm for LaTeX document.
        
        Args:
            algorithm: Algorithm to format
            
        Returns:
            LaTeX formatted string
        """
        lines = []
        
        # Use algorithm environment
        lines.append("\\begin{algorithm}[H]")
        lines.append(f"\\caption{{{algorithm.name}}}")
        lines.append("\\begin{algorithmic}[1]")
        
        # Add input/output
        if algorithm.input_spec:
            lines.append(f"\\Require {algorithm.input_spec}")
        if algorithm.output_spec:
            lines.append(f"\\Ensure {algorithm.output_spec}")
        
        # Add pseudocode
        if algorithm.pseudocode:
            # Convert pseudocode to algorithmic format
            for line in algorithm.pseudocode.split("\n"):
                line = line.strip()
                if not line:
                    continue
                
                # Handle common pseudocode patterns
                if line.lower().startswith("for "):
                    lines.append(f"\\For{{{line[4:]}}}")
                elif line.lower().startswith("while "):
                    lines.append(f"\\While{{{line[6:]}}}")
                elif line.lower().startswith("if "):
                    lines.append(f"\\If{{{line[3:]}}}")
                elif line.lower().startswith("else if "):
                    lines.append(f"\\ElsIf{{{line[8:]}}}")
                elif line.lower() == "else":
                    lines.append("\\Else")
                elif line.lower().startswith("return "):
                    lines.append(f"\\Return {line[7:]}")
                elif line.lower() in ["end for", "endfor"]:
                    lines.append("\\EndFor")
                elif line.lower() in ["end while", "endwhile"]:
                    lines.append("\\EndWhile")
                elif line.lower() in ["end if", "endif"]:
                    lines.append("\\EndIf")
                else:
                    lines.append(f"\\State {line}")
        
        lines.append("\\end{algorithmic}")
        lines.append("\\end{algorithm}")
        
        # Add complexity note
        if algorithm.complexity_str:
            lines.append(f"\\textbf{{Complexity:}} {algorithm.complexity_str}")
        
        return "\n".join(lines)
    
    def format_pseudocode_simple(self, algorithm: Algorithm) -> str:
        """
        Format algorithm as simple pseudocode block.
        
        Args:
            algorithm: Algorithm to format
            
        Returns:
            Formatted pseudocode string
        """
        lines = [
            f"Algorithm: {algorithm.name}",
            f"Input: {algorithm.input_spec}" if algorithm.input_spec else None,
            f"Output: {algorithm.output_spec}" if algorithm.output_spec else None,
            "",
            algorithm.pseudocode,
        ]
        
        if algorithm.complexity_str:
            lines.extend(["", f"Complexity: {algorithm.complexity_str}"])
        
        return "\n".join(line for line in lines if line is not None)



"""
Vector database abstraction layer.

Provides a unified interface for different vector store backends:
- Chroma
- Qdrant  
- Weaviate
"""

import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np

from .logging import get_logger

logger = get_logger(__name__)


@dataclass
class Document:
    """Represents a document chunk with metadata."""
    id: str
    content: str
    embedding: Optional[List[float]] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    @classmethod
    def create(
        cls,
        content: str,
        metadata: Optional[Dict[str, Any]] = None,
        embedding: Optional[List[float]] = None
    ) -> "Document":
        """Create a new document with auto-generated ID."""
        return cls(
            id=str(uuid.uuid4()),
            content=content,
            embedding=embedding,
            metadata=metadata or {}
        )


@dataclass
class SearchResult:
    """Represents a search result from the vector store."""
    document: Document
    score: float
    
    @property
    def content(self) -> str:
        return self.document.content
    
    @property
    def metadata(self) -> Dict[str, Any]:
        return self.document.metadata


class VectorStore(ABC):
    """Abstract base class for vector store implementations."""
    
    @abstractmethod
    def add_documents(self, documents: List[Document]) -> List[str]:
        """
        Add documents to the vector store.
        
        Args:
            documents: List of documents to add
            
        Returns:
            List of document IDs
        """
        pass
    
    @abstractmethod
    def search(
        self,
        query_embedding: List[float],
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[SearchResult]:
        """
        Search for similar documents.
        
        Args:
            query_embedding: Query vector
            top_k: Number of results to return
            filter_metadata: Optional metadata filters
            
        Returns:
            List of search results
        """
        pass
    
    @abstractmethod
    def delete(self, document_ids: List[str]) -> None:
        """Delete documents by ID."""
        pass
    
    @abstractmethod
    def clear(self) -> None:
        """Clear all documents from the store."""
        pass
    
    @abstractmethod
    def count(self) -> int:
        """Get the number of documents in the store."""
        pass
    
    def search_by_text(
        self,
        query: str,
        embed_fn: callable,
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[SearchResult]:
        """
        Search using text query (requires embedding function).
        
        Args:
            query: Text query
            embed_fn: Function to embed text
            top_k: Number of results
            filter_metadata: Optional filters
            
        Returns:
            List of search results
        """
        query_embedding = embed_fn(query)
        return self.search(query_embedding, top_k, filter_metadata)


class ChromaStore(VectorStore):
    """Chroma vector store implementation."""
    
    def __init__(
        self,
        collection_name: str = "readanybook",
        persist_directory: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        distance_metric: str = "cosine"
    ):
        """
        Initialize Chroma store.
        
        Args:
            collection_name: Name of the collection
            persist_directory: Directory for persistent storage (local mode)
            host: Chroma server host (server mode)
            port: Chroma server port (server mode)
            distance_metric: Distance metric (cosine, l2, ip)
        """
        try:
            import chromadb
            from chromadb.config import Settings as ChromaSettings
        except ImportError:
            raise ImportError("chromadb is required. Install with: pip install chromadb")
        
        self.collection_name = collection_name
        self.distance_metric = distance_metric
        
        # Initialize client
        if host and port:
            # Server mode
            self.client = chromadb.HttpClient(host=host, port=port)
            logger.info(f"Connected to Chroma server at {host}:{port}")
        elif persist_directory:
            # Persistent local mode
            Path(persist_directory).mkdir(parents=True, exist_ok=True)
            self.client = chromadb.PersistentClient(path=persist_directory)
            logger.info(f"Initialized persistent Chroma at {persist_directory}")
        else:
            # In-memory mode
            self.client = chromadb.Client()
            logger.info("Initialized in-memory Chroma")
        
        # Get or create collection
        self.collection = self.client.get_or_create_collection(
            name=collection_name,
            metadata={"hnsw:space": distance_metric}
        )
    
    def add_documents(self, documents: List[Document]) -> List[str]:
        """Add documents to Chroma."""
        if not documents:
            return []
        
        ids = [doc.id for doc in documents]
        contents = [doc.content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        embeddings = [doc.embedding for doc in documents if doc.embedding]
        
        add_kwargs = {
            "ids": ids,
            "documents": contents,
            "metadatas": metadatas
        }
        
        if embeddings and len(embeddings) == len(documents):
            add_kwargs["embeddings"] = embeddings
        
        self.collection.add(**add_kwargs)
        logger.info(f"Added {len(documents)} documents to Chroma")
        
        return ids
    
    def search(
        self,
        query_embedding: List[float],
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[SearchResult]:
        """Search Chroma for similar documents."""
        query_kwargs = {
            "query_embeddings": [query_embedding],
            "n_results": top_k,
            "include": ["documents", "metadatas", "distances", "embeddings"]
        }
        
        if filter_metadata:
            query_kwargs["where"] = filter_metadata
        
        results = self.collection.query(**query_kwargs)
        
        search_results = []
        if results["ids"] and results["ids"][0]:
            for i, doc_id in enumerate(results["ids"][0]):
                doc = Document(
                    id=doc_id,
                    content=results["documents"][0][i] if results["documents"] else "",
                    metadata=results["metadatas"][0][i] if results["metadatas"] else {},
                    embedding=results["embeddings"][0][i] if results.get("embeddings") else None
                )
                # Chroma returns distances, convert to similarity score
                distance = results["distances"][0][i] if results["distances"] else 0
                score = 1 - distance if self.distance_metric == "cosine" else 1 / (1 + distance)
                search_results.append(SearchResult(document=doc, score=score))
        
        return search_results
    
    def delete(self, document_ids: List[str]) -> None:
        """Delete documents from Chroma."""
        if document_ids:
            self.collection.delete(ids=document_ids)
            logger.info(f"Deleted {len(document_ids)} documents from Chroma")
    
    def clear(self) -> None:
        """Clear all documents from collection."""
        self.client.delete_collection(self.collection_name)
        self.collection = self.client.create_collection(
            name=self.collection_name,
            metadata={"hnsw:space": self.distance_metric}
        )
        logger.info("Cleared Chroma collection")
    
    def count(self) -> int:
        """Get document count."""
        return self.collection.count()


class QdrantStore(VectorStore):
    """Qdrant vector store implementation."""
    
    def __init__(
        self,
        collection_name: str = "readanybook",
        host: str = "localhost",
        port: int = 6333,
        grpc_port: int = 6334,
        vector_size: int = 384,
        distance_metric: str = "cosine",
        persist_directory: Optional[str] = None
    ):
        """
        Initialize Qdrant store.
        
        Args:
            collection_name: Name of the collection
            host: Qdrant server host
            port: Qdrant REST API port
            grpc_port: Qdrant gRPC port
            vector_size: Dimension of vectors
            distance_metric: Distance metric
            persist_directory: Path for local storage (if no server)
        """
        try:
            from qdrant_client import QdrantClient
            from qdrant_client.models import Distance, VectorParams
        except ImportError:
            raise ImportError("qdrant-client is required. Install with: pip install qdrant-client")
        
        self.collection_name = collection_name
        self.vector_size = vector_size
        
        # Map distance metric
        distance_map = {
            "cosine": Distance.COSINE,
            "l2": Distance.EUCLID,
            "ip": Distance.DOT
        }
        self.distance = distance_map.get(distance_metric, Distance.COSINE)
        
        # Initialize client
        if persist_directory:
            Path(persist_directory).mkdir(parents=True, exist_ok=True)
            self.client = QdrantClient(path=persist_directory)
            logger.info(f"Initialized local Qdrant at {persist_directory}")
        else:
            self.client = QdrantClient(host=host, port=port, grpc_port=grpc_port)
            logger.info(f"Connected to Qdrant server at {host}:{port}")
        
        # Create collection if not exists
        collections = self.client.get_collections().collections
        if not any(c.name == collection_name for c in collections):
            self.client.create_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=vector_size, distance=self.distance)
            )
            logger.info(f"Created Qdrant collection: {collection_name}")
    
    def add_documents(self, documents: List[Document]) -> List[str]:
        """Add documents to Qdrant."""
        if not documents:
            return []
        
        try:
            from qdrant_client.models import PointStruct
        except ImportError:
            raise ImportError("qdrant-client is required")
        
        points = []
        for doc in documents:
            if doc.embedding is None:
                continue
            
            payload = {
                "content": doc.content,
                **doc.metadata
            }
            
            points.append(PointStruct(
                id=doc.id,
                vector=doc.embedding,
                payload=payload
            ))
        
        if points:
            self.client.upsert(collection_name=self.collection_name, points=points)
            logger.info(f"Added {len(points)} documents to Qdrant")
        
        return [doc.id for doc in documents]
    
    def search(
        self,
        query_embedding: List[float],
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[SearchResult]:
        """Search Qdrant for similar documents."""
        search_params = {
            "collection_name": self.collection_name,
            "query_vector": query_embedding,
            "limit": top_k,
            "with_payload": True,
            "with_vectors": True
        }
        
        if filter_metadata:
            from qdrant_client.models import Filter, FieldCondition, MatchValue
            conditions = [
                FieldCondition(key=k, match=MatchValue(value=v))
                for k, v in filter_metadata.items()
            ]
            search_params["query_filter"] = Filter(must=conditions)
        
        results = self.client.search(**search_params)
        
        search_results = []
        for hit in results:
            content = hit.payload.pop("content", "")
            doc = Document(
                id=str(hit.id),
                content=content,
                metadata=hit.payload,
                embedding=hit.vector
            )
            search_results.append(SearchResult(document=doc, score=hit.score))
        
        return search_results
    
    def delete(self, document_ids: List[str]) -> None:
        """Delete documents from Qdrant."""
        if document_ids:
            from qdrant_client.models import PointIdsList
            self.client.delete(
                collection_name=self.collection_name,
                points_selector=PointIdsList(points=document_ids)
            )
            logger.info(f"Deleted {len(document_ids)} documents from Qdrant")
    
    def clear(self) -> None:
        """Clear all documents from collection."""
        from qdrant_client.models import VectorParams
        
        self.client.delete_collection(self.collection_name)
        self.client.create_collection(
            collection_name=self.collection_name,
            vectors_config=VectorParams(size=self.vector_size, distance=self.distance)
        )
        logger.info("Cleared Qdrant collection")
    
    def count(self) -> int:
        """Get document count."""
        info = self.client.get_collection(self.collection_name)
        return info.points_count


class WeaviateStore(VectorStore):
    """Weaviate vector store implementation."""
    
    def __init__(
        self,
        collection_name: str = "ReadAnyBook",
        host: str = "localhost",
        port: int = 8080,
        distance_metric: str = "cosine"
    ):
        """
        Initialize Weaviate store.
        
        Args:
            collection_name: Name of the class/collection
            host: Weaviate server host
            port: Weaviate server port
            distance_metric: Distance metric
        """
        try:
            import weaviate
        except ImportError:
            raise ImportError("weaviate-client is required. Install with: pip install weaviate-client")
        
        self.collection_name = collection_name
        
        self.client = weaviate.Client(f"http://{host}:{port}")
        logger.info(f"Connected to Weaviate at {host}:{port}")
        
        # Create schema if not exists
        if not self.client.schema.exists(collection_name):
            class_schema = {
                "class": collection_name,
                "vectorizer": "none",
                "properties": [
                    {"name": "content", "dataType": ["text"]},
                    {"name": "doc_id", "dataType": ["string"]},
                    {"name": "metadata", "dataType": ["text"]}
                ]
            }
            self.client.schema.create_class(class_schema)
            logger.info(f"Created Weaviate class: {collection_name}")
    
    def add_documents(self, documents: List[Document]) -> List[str]:
        """Add documents to Weaviate."""
        import json
        
        if not documents:
            return []
        
        with self.client.batch as batch:
            for doc in documents:
                if doc.embedding is None:
                    continue
                
                properties = {
                    "content": doc.content,
                    "doc_id": doc.id,
                    "metadata": json.dumps(doc.metadata)
                }
                
                batch.add_data_object(
                    properties,
                    self.collection_name,
                    vector=doc.embedding,
                    uuid=doc.id
                )
        
        logger.info(f"Added {len(documents)} documents to Weaviate")
        return [doc.id for doc in documents]
    
    def search(
        self,
        query_embedding: List[float],
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[SearchResult]:
        """Search Weaviate for similar documents."""
        import json
        
        query = (
            self.client.query
            .get(self.collection_name, ["content", "doc_id", "metadata"])
            .with_near_vector({"vector": query_embedding})
            .with_limit(top_k)
            .with_additional(["certainty", "vector"])
        )
        
        results = query.do()
        
        search_results = []
        if results.get("data", {}).get("Get", {}).get(self.collection_name):
            for item in results["data"]["Get"][self.collection_name]:
                metadata = json.loads(item.get("metadata", "{}"))
                doc = Document(
                    id=item.get("doc_id", ""),
                    content=item.get("content", ""),
                    metadata=metadata,
                    embedding=item.get("_additional", {}).get("vector")
                )
                score = item.get("_additional", {}).get("certainty", 0)
                search_results.append(SearchResult(document=doc, score=score))
        
        return search_results
    
    def delete(self, document_ids: List[str]) -> None:
        """Delete documents from Weaviate."""
        for doc_id in document_ids:
            self.client.data_object.delete(doc_id, self.collection_name)
        logger.info(f"Deleted {len(document_ids)} documents from Weaviate")
    
    def clear(self) -> None:
        """Clear all documents from collection."""
        self.client.schema.delete_class(self.collection_name)
        # Recreate class
        class_schema = {
            "class": self.collection_name,
            "vectorizer": "none",
            "properties": [
                {"name": "content", "dataType": ["text"]},
                {"name": "doc_id", "dataType": ["string"]},
                {"name": "metadata", "dataType": ["text"]}
            ]
        }
        self.client.schema.create_class(class_schema)
        logger.info("Cleared Weaviate collection")
    
    def count(self) -> int:
        """Get document count."""
        result = (
            self.client.query
            .aggregate(self.collection_name)
            .with_meta_count()
            .do()
        )
        return result.get("data", {}).get("Aggregate", {}).get(self.collection_name, [{}])[0].get("meta", {}).get("count", 0)


def create_vector_store(settings) -> VectorStore:
    """
    Factory function to create a vector store from settings.
    
    Args:
        settings: VectorDBSettings or Settings object
        
    Returns:
        VectorStore instance
    """
    if hasattr(settings, "vectordb"):
        vdb_settings = settings.vectordb
    else:
        vdb_settings = settings
    
    backend = vdb_settings.backend
    
    if backend == "chroma":
        return ChromaStore(
            collection_name=vdb_settings.collection_name,
            persist_directory=vdb_settings.persist_directory,
            host=vdb_settings.chroma.host,
            port=vdb_settings.chroma.port,
            distance_metric=vdb_settings.distance_metric
        )
    elif backend == "qdrant":
        return QdrantStore(
            collection_name=vdb_settings.collection_name,
            host=vdb_settings.qdrant.host,
            port=vdb_settings.qdrant.port,
            grpc_port=vdb_settings.qdrant.grpc_port,
            distance_metric=vdb_settings.distance_metric,
            persist_directory=vdb_settings.persist_directory
        )
    elif backend == "weaviate":
        return WeaviateStore(
            collection_name=vdb_settings.collection_name,
            host=vdb_settings.weaviate.host,
            port=vdb_settings.weaviate.port,
            distance_metric=vdb_settings.distance_metric
        )
    else:
        raise ValueError(f"Unknown vector store backend: {backend}")



"""
LaTeX templates for cheat sheet generation.
"""

from pathlib import Path

TEMPLATES_DIR = Path(__file__).parent


def get_template(name: str) -> str:
    """
    Get a template by name.
    
    Args:
        name: Template name (without .tex extension)
        
    Returns:
        Template string
    """
    template_path = TEMPLATES_DIR / f"{name}.tex"
    if template_path.exists():
        return template_path.read_text(encoding='utf-8')
    
    raise ValueError(f"Template not found: {name}")


def list_templates():
    """List available templates."""
    return [p.stem for p in TEMPLATES_DIR.glob("*.tex")]


__all__ = ["get_template", "list_templates", "TEMPLATES_DIR"]



"""
Retrieval service for RAG pipeline.

Supports:
- Dense retrieval (embedding-based)
- Sparse retrieval (BM25)
- Hybrid retrieval (combination)
- Reranking
"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from .chunking import Chunk, ChunkType
from .indexing import EmbeddingModel
from ..infra.vectordb import VectorStore, SearchResult
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class RetrievalResult:
    """Result from retrieval with relevance score."""
    content: str
    score: float
    chunk_type: ChunkType
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    # Source information
    document_id: Optional[str] = None
    section_hierarchy: List[str] = field(default_factory=list)
    page_numbers: List[int] = field(default_factory=list)
    
    @classmethod
    def from_search_result(cls, result: SearchResult) -> "RetrievalResult":
        """Create from vector store search result."""
        metadata = result.metadata.copy()
        
        # Parse chunk type
        chunk_type_str = metadata.pop("chunk_type", "mixed")
        try:
            chunk_type = ChunkType(chunk_type_str)
        except ValueError:
            chunk_type = ChunkType.MIXED
        
        # Parse section hierarchy
        section_str = metadata.pop("section_hierarchy", "")
        section_hierarchy = section_str.split("|") if section_str else []
        
        # Parse page numbers
        pages_str = metadata.pop("page_numbers", "")
        page_numbers = [int(p) for p in pages_str.split(",") if p.strip().isdigit()]
        
        document_id = metadata.pop("document_id", None)
        
        return cls(
            content=result.content,
            score=result.score,
            chunk_type=chunk_type,
            metadata=metadata,
            document_id=document_id,
            section_hierarchy=section_hierarchy,
            page_numbers=page_numbers
        )


class Retriever(ABC):
    """Abstract base class for retrievers."""
    
    @abstractmethod
    def retrieve(
        self,
        query: str,
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[RetrievalResult]:
        """
        Retrieve relevant documents for a query.
        
        Args:
            query: Search query
            top_k: Number of results to return
            filter_metadata: Optional metadata filters
            
        Returns:
            List of retrieval results
        """
        pass


class DenseRetriever(Retriever):
    """Dense retriever using embeddings."""
    
    def __init__(
        self,
        vector_store: VectorStore,
        embedding_model: EmbeddingModel
    ):
        """
        Initialize dense retriever.
        
        Args:
            vector_store: Vector store for search
            embedding_model: Model for query embedding
        """
        self.vector_store = vector_store
        self.embedding_model = embedding_model
    
    @trace_function(name="dense_retrieve")
    def retrieve(
        self,
        query: str,
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[RetrievalResult]:
        """Retrieve using dense embeddings."""
        # Embed query
        query_embedding = self.embedding_model.embed(query)
        
        # Search vector store
        results = self.vector_store.search(
            query_embedding=query_embedding,
            top_k=top_k,
            filter_metadata=filter_metadata
        )
        
        # Convert to retrieval results
        return [RetrievalResult.from_search_result(r) for r in results]


class BM25Retriever(Retriever):
    """Sparse retriever using BM25."""
    
    def __init__(
        self,
        documents: Optional[List[Dict[str, Any]]] = None,
        k1: float = 1.5,
        b: float = 0.75
    ):
        """
        Initialize BM25 retriever.
        
        Args:
            documents: List of documents with 'content' and 'metadata'
            k1: BM25 k1 parameter
            b: BM25 b parameter
        """
        self.k1 = k1
        self.b = b
        self._documents = documents or []
        self._bm25 = None
        self._tokenized_corpus = None
        
        if documents:
            self._build_index(documents)
    
    def _build_index(self, documents: List[Dict[str, Any]]) -> None:
        """Build BM25 index from documents."""
        try:
            from rank_bm25 import BM25Okapi
        except ImportError:
            raise ImportError("rank_bm25 is required. Install with: pip install rank-bm25")
        
        self._documents = documents
        self._tokenized_corpus = [
            self._tokenize(doc.get("content", ""))
            for doc in documents
        ]
        
        self._bm25 = BM25Okapi(
            self._tokenized_corpus,
            k1=self.k1,
            b=self.b
        )
        
        logger.info(f"Built BM25 index with {len(documents)} documents")
    
    def _tokenize(self, text: str) -> List[str]:
        """Simple tokenization."""
        # Convert to lowercase and split on non-alphanumeric
        tokens = re.findall(r'\b\w+\b', text.lower())
        return tokens
    
    def add_documents(self, documents: List[Dict[str, Any]]) -> None:
        """Add documents and rebuild index."""
        self._build_index(self._documents + documents)
    
    @trace_function(name="bm25_retrieve")
    def retrieve(
        self,
        query: str,
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[RetrievalResult]:
        """Retrieve using BM25."""
        if self._bm25 is None:
            logger.warning("BM25 index not built")
            return []
        
        # Tokenize query
        query_tokens = self._tokenize(query)
        
        # Get BM25 scores
        scores = self._bm25.get_scores(query_tokens)
        
        # Get top-k indices
        top_indices = sorted(
            range(len(scores)),
            key=lambda i: scores[i],
            reverse=True
        )[:top_k]
        
        # Build results
        results = []
        for idx in top_indices:
            if scores[idx] > 0:
                doc = self._documents[idx]
                metadata = doc.get("metadata", {}).copy()
                
                # Parse chunk type
                chunk_type_str = metadata.pop("chunk_type", "mixed")
                try:
                    chunk_type = ChunkType(chunk_type_str)
                except ValueError:
                    chunk_type = ChunkType.MIXED
                
                results.append(RetrievalResult(
                    content=doc.get("content", ""),
                    score=scores[idx],
                    chunk_type=chunk_type,
                    metadata=metadata
                ))
        
        return results


class HybridRetriever(Retriever):
    """
    Hybrid retriever combining dense and sparse retrieval.
    
    Uses reciprocal rank fusion (RRF) to combine results.
    """
    
    def __init__(
        self,
        dense_retriever: DenseRetriever,
        sparse_retriever: BM25Retriever,
        dense_weight: float = 0.5,
        sparse_weight: float = 0.5,
        rrf_k: int = 60
    ):
        """
        Initialize hybrid retriever.
        
        Args:
            dense_retriever: Dense retriever instance
            sparse_retriever: Sparse retriever instance
            dense_weight: Weight for dense results
            sparse_weight: Weight for sparse results
            rrf_k: RRF constant (higher = more emphasis on lower ranks)
        """
        self.dense_retriever = dense_retriever
        self.sparse_retriever = sparse_retriever
        self.dense_weight = dense_weight
        self.sparse_weight = sparse_weight
        self.rrf_k = rrf_k
    
    @trace_function(name="hybrid_retrieve")
    def retrieve(
        self,
        query: str,
        top_k: int = 10,
        filter_metadata: Optional[Dict[str, Any]] = None
    ) -> List[RetrievalResult]:
        """Retrieve using hybrid approach."""
        # Get results from both retrievers
        dense_results = self.dense_retriever.retrieve(query, top_k * 2, filter_metadata)
        sparse_results = self.sparse_retriever.retrieve(query, top_k * 2, filter_metadata)
        
        # Compute RRF scores
        content_to_result: Dict[str, RetrievalResult] = {}
        content_scores: Dict[str, float] = {}
        
        # Process dense results
        for rank, result in enumerate(dense_results):
            content_key = result.content[:500]  # Use first 500 chars as key
            rrf_score = self.dense_weight / (self.rrf_k + rank + 1)
            
            content_scores[content_key] = content_scores.get(content_key, 0) + rrf_score
            if content_key not in content_to_result:
                content_to_result[content_key] = result
        
        # Process sparse results
        for rank, result in enumerate(sparse_results):
            content_key = result.content[:500]
            rrf_score = self.sparse_weight / (self.rrf_k + rank + 1)
            
            content_scores[content_key] = content_scores.get(content_key, 0) + rrf_score
            if content_key not in content_to_result:
                content_to_result[content_key] = result
        
        # Sort by combined score and return top-k
        sorted_keys = sorted(
            content_scores.keys(),
            key=lambda k: content_scores[k],
            reverse=True
        )[:top_k]
        
        results = []
        for key in sorted_keys:
            result = content_to_result[key]
            result.score = content_scores[key]
            results.append(result)
        
        return results


class Reranker:
    """
    Cross-encoder reranker for improving retrieval quality.
    """
    
    def __init__(
        self,
        model_name: str = "cross-encoder/ms-marco-MiniLM-L-6-v2",
        device: str = "auto"
    ):
        """
        Initialize reranker.
        
        Args:
            model_name: Cross-encoder model name
            device: Device to use
        """
        self.model_name = model_name
        self.device = self._resolve_device(device)
        self._model = None
    
    def _resolve_device(self, device: str) -> str:
        """Resolve device string."""
        if device != "auto":
            return device
        
        try:
            import torch
            if torch.cuda.is_available():
                return "cuda"
            elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
                return "mps"
        except ImportError:
            pass
        
        return "cpu"
    
    def _load_model(self):
        """Lazy load the reranker model."""
        if self._model is not None:
            return
        
        try:
            from sentence_transformers import CrossEncoder
            self._model = CrossEncoder(self.model_name, device=self.device)
            logger.info(f"Loaded reranker model: {self.model_name}")
        except ImportError:
            raise ImportError(
                "sentence-transformers is required for reranking. "
                "Install with: pip install sentence-transformers"
            )
    
    @trace_function(name="rerank")
    def rerank(
        self,
        query: str,
        results: List[RetrievalResult],
        top_k: Optional[int] = None
    ) -> List[RetrievalResult]:
        """
        Rerank results using cross-encoder.
        
        Args:
            query: Original query
            results: Results to rerank
            top_k: Number of results to return (all if None)
            
        Returns:
            Reranked results
        """
        if not results:
            return []
        
        self._load_model()
        
        # Create query-document pairs
        pairs = [(query, result.content) for result in results]
        
        # Get reranking scores
        scores = self._model.predict(pairs)
        
        # Update scores and sort
        for result, score in zip(results, scores):
            result.score = float(score)
        
        sorted_results = sorted(results, key=lambda r: r.score, reverse=True)
        
        if top_k:
            return sorted_results[:top_k]
        return sorted_results


class RetrievalService:
    """
    Main retrieval service orchestrating different retrieval strategies.
    """
    
    def __init__(
        self,
        vector_store: VectorStore,
        embedding_model: EmbeddingModel,
        settings=None
    ):
        """
        Initialize retrieval service.
        
        Args:
            vector_store: Vector store instance
            embedding_model: Embedding model instance
            settings: Optional RetrievalSettings
        """
        self.vector_store = vector_store
        self.embedding_model = embedding_model
        self.settings = settings
        
        # Initialize retrievers
        self.dense_retriever = DenseRetriever(vector_store, embedding_model)
        self.sparse_retriever = None  # Initialized when documents are added
        self.hybrid_retriever = None
        
        # Initialize reranker if enabled
        self.reranker = None
        if settings and settings.rerank:
            self.reranker = Reranker(
                model_name=settings.rerank_model,
                device=embedding_model.device if embedding_model else "auto"
            )
        
        # Settings
        self.strategy = settings.strategy if settings else "dense"
        self.top_k = settings.top_k if settings else 10
        self.rerank_top_k = settings.rerank_top_k if settings else 5
    
    def build_sparse_index(self, documents: List[Dict[str, Any]]) -> None:
        """
        Build sparse index for BM25 retrieval.
        
        Args:
            documents: List of documents with 'content' and 'metadata'
        """
        bm25_settings = self.settings.bm25 if self.settings else None
        
        self.sparse_retriever = BM25Retriever(
            documents=documents,
            k1=bm25_settings.k1 if bm25_settings else 1.5,
            b=bm25_settings.b if bm25_settings else 0.75
        )
        
        # Create hybrid retriever
        self.hybrid_retriever = HybridRetriever(
            dense_retriever=self.dense_retriever,
            sparse_retriever=self.sparse_retriever
        )
    
    @trace_function(name="retrieve")
    def retrieve(
        self,
        query: str,
        top_k: Optional[int] = None,
        strategy: Optional[str] = None,
        filter_metadata: Optional[Dict[str, Any]] = None,
        rerank: Optional[bool] = None
    ) -> List[RetrievalResult]:
        """
        Retrieve relevant documents for a query.
        
        Args:
            query: Search query
            top_k: Number of results (uses default if not specified)
            strategy: Retrieval strategy ("dense", "sparse", "hybrid")
            filter_metadata: Optional metadata filters
            rerank: Whether to rerank results
            
        Returns:
            List of retrieval results
        """
        top_k = top_k or self.top_k
        strategy = strategy or self.strategy
        rerank = rerank if rerank is not None else (self.reranker is not None)
        
        # Select retriever
        if strategy == "hybrid" and self.hybrid_retriever:
            retriever = self.hybrid_retriever
        elif strategy == "sparse" and self.sparse_retriever:
            retriever = self.sparse_retriever
        else:
            retriever = self.dense_retriever
        
        # Retrieve
        results = retriever.retrieve(
            query=query,
            top_k=top_k * 2 if rerank else top_k,  # Get more for reranking
            filter_metadata=filter_metadata
        )
        
        # Rerank if enabled
        if rerank and self.reranker and results:
            results = self.reranker.rerank(
                query=query,
                results=results,
                top_k=self.rerank_top_k
            )
        
        return results[:top_k]
    
    def retrieve_by_type(
        self,
        query: str,
        chunk_type: ChunkType,
        top_k: Optional[int] = None
    ) -> List[RetrievalResult]:
        """
        Retrieve documents of a specific type.
        
        Args:
            query: Search query
            chunk_type: Type of chunks to retrieve
            top_k: Number of results
            
        Returns:
            Filtered retrieval results
        """
        results = self.retrieve(
            query=query,
            top_k=(top_k or self.top_k) * 2,  # Get more to filter
            filter_metadata={"chunk_type": chunk_type.value}
        )
        
        # Filter by type (in case filter not supported)
        filtered = [r for r in results if r.chunk_type == chunk_type]
        return filtered[:top_k or self.top_k]
    
    def retrieve_math(self, query: str, top_k: int = 10) -> List[RetrievalResult]:
        """Retrieve math-related content."""
        return self.retrieve_by_type(query, ChunkType.MATH, top_k)
    
    def retrieve_code(self, query: str, top_k: int = 10) -> List[RetrievalResult]:
        """Retrieve code-related content."""
        return self.retrieve_by_type(query, ChunkType.CODE, top_k)
    
    def retrieve_narrative(self, query: str, top_k: int = 10) -> List[RetrievalResult]:
        """Retrieve narrative/explanatory content."""
        return self.retrieve_by_type(query, ChunkType.NARRATIVE, top_k)



"""
Distributed tracing support using OpenTelemetry.

Provides:
- Span creation for operations
- Latency and throughput tracking
- Integration with observability backends
"""

import functools
import time
from contextlib import contextmanager
from typing import Any, Callable, Dict, Optional, TypeVar

from .logging import get_logger

logger = get_logger(__name__)

# Type variable for generic functions
F = TypeVar("F", bound=Callable[..., Any])

# Global tracer instance
_tracer = None
_enabled = False


class Span:
    """Represents a tracing span."""
    
    def __init__(self, name: str, attributes: Optional[Dict[str, Any]] = None):
        self.name = name
        self.attributes = attributes or {}
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self._otel_span = None
    
    def __enter__(self) -> "Span":
        self.start_time = time.perf_counter()
        if _enabled and _tracer:
            self._otel_span = _tracer.start_span(self.name)
            for key, value in self.attributes.items():
                self._otel_span.set_attribute(key, value)
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.perf_counter()
        duration_ms = (self.end_time - self.start_time) * 1000
        
        if self._otel_span:
            self._otel_span.set_attribute("duration_ms", duration_ms)
            if exc_type:
                self._otel_span.set_attribute("error", True)
                self._otel_span.set_attribute("error.type", exc_type.__name__)
                self._otel_span.set_attribute("error.message", str(exc_val))
            self._otel_span.end()
        
        logger.debug(
            f"Span '{self.name}' completed",
            additional={
                "duration_ms": duration_ms,
                "attributes": self.attributes,
                "error": exc_type is not None
            }
        )
    
    def set_attribute(self, key: str, value: Any) -> None:
        """Set a span attribute."""
        self.attributes[key] = value
        if self._otel_span:
            self._otel_span.set_attribute(key, value)
    
    def add_event(self, name: str, attributes: Optional[Dict[str, Any]] = None) -> None:
        """Add an event to the span."""
        if self._otel_span:
            self._otel_span.add_event(name, attributes or {})
    
    @property
    def duration_ms(self) -> Optional[float]:
        """Get span duration in milliseconds."""
        if self.start_time and self.end_time:
            return (self.end_time - self.start_time) * 1000
        return None


class Tracer:
    """Main tracer class for creating spans."""
    
    def __init__(
        self,
        service_name: str = "readanybook",
        endpoint: Optional[str] = None,
        enabled: bool = True
    ):
        """
        Initialize the tracer.
        
        Args:
            service_name: Name of the service for tracing
            endpoint: OTLP endpoint URL
            enabled: Whether tracing is enabled
        """
        global _tracer, _enabled
        
        self.service_name = service_name
        self.endpoint = endpoint
        _enabled = enabled
        
        if enabled and endpoint:
            self._setup_opentelemetry()
        else:
            _tracer = None
            logger.info("Tracing disabled or no endpoint configured")
    
    def _setup_opentelemetry(self) -> None:
        """Set up OpenTelemetry tracing."""
        global _tracer
        
        try:
            from opentelemetry import trace
            from opentelemetry.sdk.trace import TracerProvider
            from opentelemetry.sdk.trace.export import BatchSpanProcessor
            from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
            from opentelemetry.sdk.resources import Resource
            
            resource = Resource.create({
                "service.name": self.service_name,
                "service.version": "0.1.0"
            })
            
            provider = TracerProvider(resource=resource)
            
            if self.endpoint:
                exporter = OTLPSpanExporter(endpoint=self.endpoint)
                processor = BatchSpanProcessor(exporter)
                provider.add_span_processor(processor)
            
            trace.set_tracer_provider(provider)
            _tracer = trace.get_tracer(self.service_name)
            
            logger.info(f"OpenTelemetry tracing initialized for {self.service_name}")
            
        except ImportError:
            logger.warning(
                "OpenTelemetry packages not installed. "
                "Install with: pip install opentelemetry-api opentelemetry-sdk opentelemetry-exporter-otlp"
            )
            _tracer = None
    
    def start_span(self, name: str, attributes: Optional[Dict[str, Any]] = None) -> Span:
        """Start a new span."""
        return Span(name, attributes)
    
    @contextmanager
    def span(self, name: str, attributes: Optional[Dict[str, Any]] = None):
        """Context manager for creating a span."""
        with Span(name, attributes) as s:
            yield s


def trace_function(
    name: Optional[str] = None,
    attributes: Optional[Dict[str, Any]] = None
) -> Callable[[F], F]:
    """
    Decorator to trace a function.
    
    Args:
        name: Span name (defaults to function name)
        attributes: Additional span attributes
        
    Usage:
        @trace_function()
        def my_function():
            pass
            
        @trace_function(name="custom_name", attributes={"key": "value"})
        def another_function():
            pass
    """
    def decorator(func: F) -> F:
        span_name = name or func.__name__
        span_attrs = attributes or {}
        
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with Span(span_name, {**span_attrs, "function": func.__name__}):
                return func(*args, **kwargs)
        
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            with Span(span_name, {**span_attrs, "function": func.__name__}):
                return await func(*args, **kwargs)
        
        import asyncio
        if asyncio.iscoroutinefunction(func):
            return async_wrapper  # type: ignore
        return wrapper  # type: ignore
    
    return decorator


class MetricsCollector:
    """Simple metrics collection for latency and throughput."""
    
    def __init__(self):
        self._metrics: Dict[str, list] = {}
    
    def record(self, name: str, value: float) -> None:
        """Record a metric value."""
        if name not in self._metrics:
            self._metrics[name] = []
        self._metrics[name].append(value)
    
    def get_stats(self, name: str) -> Dict[str, float]:
        """Get statistics for a metric."""
        values = self._metrics.get(name, [])
        if not values:
            return {}
        
        import statistics
        return {
            "count": len(values),
            "mean": statistics.mean(values),
            "median": statistics.median(values),
            "min": min(values),
            "max": max(values),
            "stdev": statistics.stdev(values) if len(values) > 1 else 0
        }
    
    def clear(self, name: Optional[str] = None) -> None:
        """Clear metrics."""
        if name:
            self._metrics.pop(name, None)
        else:
            self._metrics.clear()


# Global metrics collector
metrics = MetricsCollector()



"""Core module - domain logic for document processing and RAG pipeline."""

from .ingestion import DocumentIngestionService, ParsedDocument
from .chunking import ChunkingService, Chunk, ChunkType
from .indexing import IndexingService
from .retrieval import RetrievalService, HybridRetriever
from .models import LLMClient, EmbeddingModel
from .pipeline import CheatSheetPipeline

__all__ = [
    "DocumentIngestionService",
    "ParsedDocument",
    "ChunkingService",
    "Chunk",
    "ChunkType",
    "IndexingService",
    "RetrievalService",
    "HybridRetriever",
    "LLMClient",
    "EmbeddingModel",
    "CheatSheetPipeline",
]



"""
Document indexing service for building and managing the vector store.

Handles:
- Embedding generation
- Vector store population
- Index management
"""

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from .chunking import Chunk, ChunkType
from .ingestion import ParsedDocument
from ..infra.vectordb import VectorStore, Document, create_vector_store
from ..infra.logging import get_logger, LogContext
from ..infra.tracing import trace_function

logger = get_logger(__name__)


class IndexingService:
    """
    Service for indexing documents into the vector store.
    
    Manages the embedding generation and storage of document chunks.
    """
    
    def __init__(
        self,
        vector_store: VectorStore,
        embedding_model: Optional["EmbeddingModel"] = None,
        settings=None
    ):
        """
        Initialize the indexing service.
        
        Args:
            vector_store: VectorStore instance
            embedding_model: EmbeddingModel for generating embeddings
            settings: Optional Settings
        """
        self.vector_store = vector_store
        self.embedding_model = embedding_model
        self.settings = settings
    
    @trace_function(name="index_chunks")
    def index_chunks(
        self,
        chunks: List[Chunk],
        batch_size: int = 32,
        show_progress: bool = True
    ) -> List[str]:
        """
        Index chunks into the vector store.
        
        Args:
            chunks: List of chunks to index
            batch_size: Batch size for embedding generation
            show_progress: Whether to show progress
            
        Returns:
            List of document IDs
        """
        if not chunks:
            logger.warning("No chunks to index")
            return []
        
        logger.info(f"Indexing {len(chunks)} chunks")
        
        # Generate embeddings if model available
        if self.embedding_model:
            contents = [chunk.content for chunk in chunks]
            embeddings = self.embedding_model.embed_batch(contents, batch_size=batch_size)
        else:
            embeddings = [None] * len(chunks)
        
        # Create documents for vector store
        documents = []
        for chunk, embedding in zip(chunks, embeddings):
            metadata = {
                "chunk_type": chunk.chunk_type.value,
                "chunk_index": chunk.chunk_index,
                "document_id": chunk.document_id or "",
                "section_hierarchy": "|".join(chunk.section_hierarchy),
                "page_numbers": ",".join(map(str, chunk.page_numbers)),
                **chunk.metadata
            }
            
            doc = Document(
                id=chunk.id,
                content=chunk.content,
                embedding=embedding,
                metadata=metadata
            )
            documents.append(doc)
        
        # Add to vector store
        doc_ids = self.vector_store.add_documents(documents)
        
        logger.info(f"Indexed {len(doc_ids)} documents")
        return doc_ids
    
    @trace_function(name="index_document")
    def index_document(
        self,
        document: ParsedDocument,
        chunking_service: "ChunkingService",
        strategy: str = "hierarchical"
    ) -> List[str]:
        """
        Index a complete document.
        
        Args:
            document: ParsedDocument to index
            chunking_service: ChunkingService for chunking
            strategy: Chunking strategy to use
            
        Returns:
            List of chunk IDs
        """
        with LogContext(document_id=document.metadata.source_path):
            # Chunk the document
            chunks = chunking_service.chunk(document, strategy=strategy)
            
            # Index the chunks
            return self.index_chunks(chunks)
    
    def delete_document(self, document_id: str) -> None:
        """
        Delete all chunks for a document.
        
        Args:
            document_id: Document ID to delete
        """
        # Note: This requires filtering support in the vector store
        # For now, we'd need to track document-chunk mappings separately
        logger.warning("Document deletion not fully implemented - requires chunk tracking")
    
    def get_stats(self) -> Dict[str, Any]:
        """Get indexing statistics."""
        return {
            "total_documents": self.vector_store.count(),
            "vector_store_backend": type(self.vector_store).__name__
        }
    
    def clear(self) -> None:
        """Clear all indexed documents."""
        self.vector_store.clear()
        logger.info("Cleared all indexed documents")


class EmbeddingModel:
    """
    Wrapper for embedding models.
    
    Supports multiple backends:
    - Sentence Transformers
    - HuggingFace Transformers
    - Custom embedding functions
    """
    
    def __init__(
        self,
        model_name: str = "BAAI/bge-small-en-v1.5",
        model_type: str = "sentence-transformers",
        device: str = "auto",
        normalize: bool = True,
        batch_size: int = 32
    ):
        """
        Initialize the embedding model.
        
        Args:
            model_name: Model name/path
            model_type: Type of model ("sentence-transformers", "huggingface")
            device: Device to use ("auto", "cpu", "cuda", "mps")
            normalize: Whether to normalize embeddings
            batch_size: Default batch size
        """
        self.model_name = model_name
        self.model_type = model_type
        self.device = self._resolve_device(device)
        self.normalize = normalize
        self.batch_size = batch_size
        
        self._model = None
        self._dimension = None
    
    def _resolve_device(self, device: str) -> str:
        """Resolve device string to actual device."""
        if device != "auto":
            return device
        
        try:
            import torch
            if torch.cuda.is_available():
                return "cuda"
            elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
                return "mps"
        except ImportError:
            pass
        
        return "cpu"
    
    def _load_model(self):
        """Lazy load the embedding model."""
        if self._model is not None:
            return
        
        logger.info(f"Loading embedding model: {self.model_name}")
        
        if self.model_type == "sentence-transformers":
            try:
                from sentence_transformers import SentenceTransformer
                self._model = SentenceTransformer(self.model_name, device=self.device)
                self._dimension = self._model.get_sentence_embedding_dimension()
            except ImportError:
                raise ImportError(
                    "sentence-transformers is required. "
                    "Install with: pip install sentence-transformers"
                )
        elif self.model_type == "huggingface":
            try:
                from transformers import AutoModel, AutoTokenizer
                import torch
                
                self._tokenizer = AutoTokenizer.from_pretrained(self.model_name)
                self._model = AutoModel.from_pretrained(self.model_name).to(self.device)
                self._model.eval()
                
                # Get dimension from model config
                self._dimension = self._model.config.hidden_size
            except ImportError:
                raise ImportError(
                    "transformers and torch are required. "
                    "Install with: pip install transformers torch"
                )
        else:
            raise ValueError(f"Unknown model type: {self.model_type}")
        
        logger.info(f"Loaded embedding model with dimension {self._dimension}")
    
    @property
    def dimension(self) -> int:
        """Get embedding dimension."""
        self._load_model()
        return self._dimension
    
    @trace_function(name="embed_text")
    def embed(self, text: str) -> List[float]:
        """
        Embed a single text.
        
        Args:
            text: Text to embed
            
        Returns:
            Embedding vector as list of floats
        """
        self._load_model()
        
        if self.model_type == "sentence-transformers":
            embedding = self._model.encode(
                text,
                normalize_embeddings=self.normalize,
                convert_to_numpy=True
            )
            return embedding.tolist()
        else:
            return self._embed_with_transformers([text])[0]
    
    @trace_function(name="embed_batch")
    def embed_batch(
        self,
        texts: List[str],
        batch_size: Optional[int] = None,
        show_progress: bool = False
    ) -> List[List[float]]:
        """
        Embed a batch of texts.
        
        Args:
            texts: List of texts to embed
            batch_size: Batch size (uses default if not specified)
            show_progress: Whether to show progress bar
            
        Returns:
            List of embedding vectors
        """
        self._load_model()
        batch_size = batch_size or self.batch_size
        
        if self.model_type == "sentence-transformers":
            embeddings = self._model.encode(
                texts,
                batch_size=batch_size,
                normalize_embeddings=self.normalize,
                show_progress_bar=show_progress,
                convert_to_numpy=True
            )
            return embeddings.tolist()
        else:
            return self._embed_with_transformers(texts, batch_size)
    
    def _embed_with_transformers(
        self,
        texts: List[str],
        batch_size: int = 32
    ) -> List[List[float]]:
        """Embed using HuggingFace transformers."""
        import torch
        import numpy as np
        
        all_embeddings = []
        
        for i in range(0, len(texts), batch_size):
            batch_texts = texts[i:i + batch_size]
            
            # Tokenize
            inputs = self._tokenizer(
                batch_texts,
                padding=True,
                truncation=True,
                max_length=512,
                return_tensors="pt"
            ).to(self.device)
            
            # Generate embeddings
            with torch.no_grad():
                outputs = self._model(**inputs)
                
                # Mean pooling
                attention_mask = inputs["attention_mask"]
                token_embeddings = outputs.last_hidden_state
                
                input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
                embeddings = torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                    input_mask_expanded.sum(1), min=1e-9
                )
                
                if self.normalize:
                    embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                
                all_embeddings.extend(embeddings.cpu().numpy().tolist())
        
        return all_embeddings


def create_embedding_model(settings) -> EmbeddingModel:
    """
    Factory function to create an embedding model from settings.
    
    Args:
        settings: EmbeddingSettings or Settings object
        
    Returns:
        EmbeddingModel instance
    """
    if hasattr(settings, "embedding"):
        emb_settings = settings.embedding
    else:
        emb_settings = settings
    
    return EmbeddingModel(
        model_name=emb_settings.model_name,
        model_type=emb_settings.model_type,
        device=emb_settings.device,
        normalize=emb_settings.normalize,
        batch_size=emb_settings.batch_size
    )



"""
Concept extraction chain for identifying and summarizing key concepts.
"""

import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from ..core.retrieval import RetrievalService, RetrievalResult
from ..core.models import LLMClient, GenerationConfig
from ..core.prompts import get_prompt
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class Concept:
    """Represents an extracted concept."""
    name: str
    definition: str = ""
    intuition: str = ""
    related: List[str] = field(default_factory=list)
    source_sections: List[str] = field(default_factory=list)
    importance_score: float = 0.0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "definition": self.definition,
            "intuition": self.intuition,
            "related": self.related,
            "source_sections": self.source_sections,
            "importance_score": self.importance_score
        }


class ConceptExtractor:
    """
    Chain for extracting concepts from documents.
    
    Uses RAG to identify and define key concepts with
    definitions and intuitive explanations.
    """
    
    def __init__(
        self,
        retrieval_service: RetrievalService,
        llm_client: LLMClient,
        max_concepts: int = 50,
        include_definitions: bool = True,
        include_intuitions: bool = True
    ):
        """
        Initialize concept extractor.
        
        Args:
            retrieval_service: Service for retrieving context
            llm_client: LLM for generation
            max_concepts: Maximum number of concepts to extract
            include_definitions: Include formal definitions
            include_intuitions: Include intuitive explanations
        """
        self.retrieval_service = retrieval_service
        self.llm_client = llm_client
        self.max_concepts = max_concepts
        self.include_definitions = include_definitions
        self.include_intuitions = include_intuitions
    
    @trace_function(name="extract_concepts")
    def extract(
        self,
        queries: Optional[List[str]] = None,
        config: Optional[GenerationConfig] = None
    ) -> List[Concept]:
        """
        Extract concepts from the indexed document.
        
        Args:
            queries: Optional specific queries for concept retrieval
            config: Generation configuration
            
        Returns:
            List of extracted concepts
        """
        config = config or GenerationConfig()
        
        # Default queries for concept extraction
        if queries is None:
            queries = [
                "key concepts definitions terminology",
                "fundamental ideas principles",
                "important terms notation",
                "core concepts main ideas",
            ]
        
        all_concepts = []
        seen_names = set()
        
        for query in queries:
            # Retrieve relevant content
            results = self.retrieval_service.retrieve_narrative(query, top_k=5)
            
            if not results:
                continue
            
            # Extract concepts from context
            concepts = self._extract_from_context(results, config)
            
            # Deduplicate
            for concept in concepts:
                if concept.name.lower() not in seen_names:
                    all_concepts.append(concept)
                    seen_names.add(concept.name.lower())
        
        # Sort by importance and limit
        all_concepts.sort(key=lambda c: c.importance_score, reverse=True)
        
        logger.info(f"Extracted {len(all_concepts)} concepts")
        return all_concepts[:self.max_concepts]
    
    def _extract_from_context(
        self,
        results: List[RetrievalResult],
        config: GenerationConfig
    ) -> List[Concept]:
        """Extract concepts from retrieval results."""
        context = "\n\n".join(r.content for r in results)
        
        # Build prompt
        prompt = get_prompt("concept_extraction")(context=context)
        
        # Generate
        result = self.llm_client.generate(prompt, config)
        
        # Parse response
        concepts = self._parse_response(result.text, results)
        
        return concepts
    
    def _parse_response(
        self,
        text: str,
        results: List[RetrievalResult]
    ) -> List[Concept]:
        """Parse concept extraction response."""
        concepts = []
        current: Dict[str, Any] = {}
        
        # Track source sections
        source_sections = list(set(
            section
            for r in results
            for section in r.section_hierarchy
            if section
        ))
        
        for line in text.split("\n"):
            line = line.strip()
            
            if line.startswith("CONCEPT:"):
                if current and "name" in current:
                    concepts.append(self._create_concept(current, source_sections))
                current = {"name": line[8:].strip()}
            elif line.startswith("DEFINITION:") and self.include_definitions:
                current["definition"] = line[11:].strip()
            elif line.startswith("INTUITION:") and self.include_intuitions:
                current["intuition"] = line[10:].strip()
            elif line.startswith("RELATED:"):
                related_str = line[8:].strip()
                current["related"] = [r.strip() for r in related_str.split(",") if r.strip()]
        
        # Don't forget the last concept
        if current and "name" in current:
            concepts.append(self._create_concept(current, source_sections))
        
        return concepts
    
    def _create_concept(
        self,
        data: Dict[str, Any],
        source_sections: List[str]
    ) -> Concept:
        """Create a Concept from parsed data."""
        # Calculate importance score based on completeness
        score = 0.0
        if data.get("definition"):
            score += 0.4
        if data.get("intuition"):
            score += 0.3
        if data.get("related"):
            score += 0.2
        if len(data.get("name", "")) > 3:
            score += 0.1
        
        return Concept(
            name=data.get("name", ""),
            definition=data.get("definition", ""),
            intuition=data.get("intuition", ""),
            related=data.get("related", []),
            source_sections=source_sections,
            importance_score=score
        )
    
    def summarize_concepts(
        self,
        concepts: List[Concept],
        max_words: int = 500,
        config: Optional[GenerationConfig] = None
    ) -> str:
        """
        Summarize concepts into a cohesive text.
        
        Args:
            concepts: Concepts to summarize
            max_words: Maximum words
            config: Generation configuration
            
        Returns:
            Summarized concept text
        """
        config = config or GenerationConfig()
        
        # Format concepts for prompt
        concept_text = "\n\n".join(
            f"- **{c.name}**: {c.definition}"
            for c in concepts
            if c.definition
        )
        
        prompt = get_prompt("concept_summary")(
            concepts=concept_text,
            max_words=max_words
        )
        
        result = self.llm_client.generate(prompt, config)
        return result.text
    
    def group_by_topic(self, concepts: List[Concept]) -> Dict[str, List[Concept]]:
        """
        Group concepts by related topics.
        
        Args:
            concepts: Concepts to group
            
        Returns:
            Dictionary mapping topic to concepts
        """
        # Simple grouping based on related concepts
        groups: Dict[str, List[Concept]] = {"General": []}
        
        for concept in concepts:
            if concept.related:
                # Use first related concept as group
                group_name = concept.related[0]
                if group_name not in groups:
                    groups[group_name] = []
                groups[group_name].append(concept)
            else:
                groups["General"].append(concept)
        
        # Remove empty groups
        return {k: v for k, v in groups.items() if v}



"""
Settings and configuration management using Pydantic.

Supports:
- YAML/TOML configuration files
- Environment variable overrides
- Profile-based configurations
"""

import os
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

import yaml
from pydantic import BaseModel, Field, field_validator
from pydantic_settings import BaseSettings


class AppSettings(BaseModel):
    """Application-level settings."""
    name: str = "ReadAnyBook"
    version: str = "0.1.0"
    log_level: str = "INFO"
    temp_dir: str = "/tmp/readanybook"


class IngestionSettings(BaseModel):
    """Document ingestion settings."""
    supported_formats: List[str] = [".pdf", ".epub", ".html", ".tex", ".md"]
    max_file_size_mb: int = 100
    extract_images: bool = False
    ocr_enabled: bool = False


class ChunkingSettings(BaseModel):
    """Text chunking settings."""
    chunk_size: int = 1000
    chunk_overlap: int = 200
    min_chunk_size: int = 100
    max_chunk_size: int = 2000
    separators: List[str] = ["\n\n", "\n", ". ", " "]
    preserve_math: bool = True
    preserve_code: bool = True


class EmbeddingSettings(BaseModel):
    """Embedding model settings."""
    model_name: str = "BAAI/bge-small-en-v1.5"
    model_type: str = "sentence-transformers"
    batch_size: int = 32
    normalize: bool = True
    device: str = "auto"


class ChromaSettings(BaseModel):
    """Chroma-specific settings."""
    host: Optional[str] = None
    port: Optional[int] = None


class QdrantSettings(BaseModel):
    """Qdrant-specific settings."""
    host: str = "localhost"
    port: int = 6333
    grpc_port: int = 6334


class WeaviateSettings(BaseModel):
    """Weaviate-specific settings."""
    host: str = "localhost"
    port: int = 8080


class VectorDBSettings(BaseModel):
    """Vector database settings."""
    backend: Literal["chroma", "qdrant", "weaviate"] = "chroma"
    persist_directory: str = "./data/vectordb"
    collection_name: str = "readanybook"
    distance_metric: str = "cosine"
    chroma: ChromaSettings = ChromaSettings()
    qdrant: QdrantSettings = QdrantSettings()
    weaviate: WeaviateSettings = WeaviateSettings()


class BM25Settings(BaseModel):
    """BM25 sparse retrieval settings."""
    k1: float = 1.5
    b: float = 0.75


class RetrievalSettings(BaseModel):
    """Retrieval settings."""
    strategy: Literal["dense", "sparse", "hybrid"] = "hybrid"
    top_k: int = 10
    rerank: bool = True
    rerank_model: str = "cross-encoder/ms-marco-MiniLM-L-6-v2"
    rerank_top_k: int = 5
    bm25: BM25Settings = BM25Settings()


class HuggingFaceSettings(BaseModel):
    """HuggingFace-specific LLM settings."""
    trust_remote_code: bool = True
    torch_dtype: str = "auto"


class VLLMSettings(BaseModel):
    """vLLM-specific settings."""
    tensor_parallel_size: int = 1


class OllamaSettings(BaseModel):
    """Ollama-specific settings."""
    host: str = "http://localhost:11434"


class OpenAISettings(BaseModel):
    """OpenAI-compatible API settings."""
    api_base: Optional[str] = None
    api_key: Optional[str] = None


class LLMSettings(BaseModel):
    """LLM settings."""
    backend: Literal["huggingface", "vllm", "ollama", "openai"] = "huggingface"
    model_name: str = "mistralai/Mistral-7B-Instruct-v0.2"
    max_tokens: int = 2048
    temperature: float = 0.7
    top_p: float = 0.9
    device: str = "auto"
    quantization: Optional[Literal["4bit", "8bit"]] = "4bit"
    huggingface: HuggingFaceSettings = HuggingFaceSettings()
    vllm: VLLMSettings = VLLMSettings()
    ollama: OllamaSettings = OllamaSettings()
    openai: OpenAISettings = OpenAISettings()


class ConceptsGenerationSettings(BaseModel):
    """Concept extraction settings."""
    max_concepts: int = 50
    include_definitions: bool = True
    include_intuitions: bool = True


class FormulasGenerationSettings(BaseModel):
    """Formula extraction settings."""
    max_formulas: int = 100
    normalize_latex: bool = True
    include_derivations: bool = False
    group_by_topic: bool = True


class ModelsGenerationSettings(BaseModel):
    """Model summarization settings."""
    include_assumptions: bool = True
    include_objectives: bool = True
    include_training: bool = True


class AlgorithmsGenerationSettings(BaseModel):
    """Algorithm synthesis settings."""
    include_pseudocode: bool = True
    include_complexity: bool = True
    include_examples: bool = True


class GenerationSettings(BaseModel):
    """Generation pipeline settings."""
    concepts: ConceptsGenerationSettings = ConceptsGenerationSettings()
    formulas: FormulasGenerationSettings = FormulasGenerationSettings()
    models: ModelsGenerationSettings = ModelsGenerationSettings()
    algorithms: AlgorithmsGenerationSettings = AlgorithmsGenerationSettings()


class MarginsSettings(BaseModel):
    """LaTeX margin settings."""
    top: str = "1.5cm"
    bottom: str = "1.5cm"
    left: str = "1.5cm"
    right: str = "1.5cm"


class LaTeXSettings(BaseModel):
    """LaTeX document settings."""
    template: str = "default"
    compiler: Literal["pdflatex", "xelatex", "lualatex", "tectonic"] = "pdflatex"
    output_format: str = "pdf"
    page_limit: int = 12
    font_size: int = 10
    paper_size: str = "a4paper"
    margins: MarginsSettings = MarginsSettings()
    columns: int = 2
    include_toc: bool = False
    include_references: bool = True


class EvaluationSettings(BaseModel):
    """Evaluation settings."""
    enabled: bool = True
    metrics: List[str] = [
        "retrieval_precision",
        "retrieval_recall",
        "answer_relevance",
        "faithfulness",
        "latex_validity",
    ]
    sample_size: int = 100


class TracingSettings(BaseModel):
    """OpenTelemetry tracing settings."""
    enabled: bool = False
    endpoint: str = "http://localhost:4317"
    service_name: str = "readanybook"


class ObservabilitySettings(BaseModel):
    """Logging and tracing settings."""
    structured_logging: bool = True
    log_format: str = "json"
    log_file: str = "./logs/readanybook.log"
    tracing: TracingSettings = TracingSettings()


class Settings(BaseSettings):
    """
    Main settings class for ReadAnyBook.
    
    Supports loading from:
    - YAML configuration files
    - Environment variables (prefixed with READANYBOOK_)
    - Programmatic overrides
    """
    
    app: AppSettings = AppSettings()
    ingestion: IngestionSettings = IngestionSettings()
    chunking: ChunkingSettings = ChunkingSettings()
    embedding: EmbeddingSettings = EmbeddingSettings()
    vectordb: VectorDBSettings = VectorDBSettings()
    retrieval: RetrievalSettings = RetrievalSettings()
    llm: LLMSettings = LLMSettings()
    generation: GenerationSettings = GenerationSettings()
    latex: LaTeXSettings = LaTeXSettings()
    evaluation: EvaluationSettings = EvaluationSettings()
    observability: ObservabilitySettings = ObservabilitySettings()
    
    class Config:
        env_prefix = "READANYBOOK_"
        env_nested_delimiter = "__"
    
    @classmethod
    def from_yaml(cls, path: Union[str, Path]) -> "Settings":
        """Load settings from a YAML file."""
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Configuration file not found: {path}")
        
        with open(path, "r") as f:
            config_dict = yaml.safe_load(f)
        
        return cls(**config_dict)
    
    @classmethod
    def from_profile(cls, profile: str, base_config: Optional[Path] = None) -> "Settings":
        """
        Load settings with a specific profile applied.
        
        Profiles can be: technical_book, math_paper, nontechnical_book
        """
        if base_config:
            settings = cls.from_yaml(base_config)
        else:
            settings = cls()
        
        # Get default config path
        default_config = Path(__file__).parent.parent / "config" / "default.yaml"
        if default_config.exists():
            with open(default_config, "r") as f:
                config_dict = yaml.safe_load(f)
            
            profiles = config_dict.get("profiles", {})
            if profile in profiles:
                profile_config = profiles[profile]
                settings = settings.apply_overrides(profile_config)
        
        return settings
    
    def apply_overrides(self, overrides: Dict[str, Any]) -> "Settings":
        """Apply configuration overrides and return a new Settings instance."""
        current_dict = self.model_dump()
        
        def deep_merge(base: dict, override: dict) -> dict:
            result = base.copy()
            for key, value in override.items():
                if key in result and isinstance(result[key], dict) and isinstance(value, dict):
                    result[key] = deep_merge(result[key], value)
                else:
                    result[key] = value
            return result
        
        merged = deep_merge(current_dict, overrides)
        return Settings(**merged)
    
    def to_yaml(self, path: Union[str, Path]) -> None:
        """Save settings to a YAML file."""
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        
        with open(path, "w") as f:
            yaml.dump(self.model_dump(), f, default_flow_style=False, sort_keys=False)


def get_default_settings() -> Settings:
    """Get default settings, optionally loading from config file."""
    default_config = Path(__file__).parent.parent / "config" / "default.yaml"
    
    if default_config.exists():
        return Settings.from_yaml(default_config)
    
    return Settings()



"""
Model summarization chain for extracting model descriptions and theories.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from ..core.retrieval import RetrievalService, RetrievalResult
from ..core.models import LLMClient, GenerationConfig
from ..core.prompts import get_prompt
from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class ModelSummary:
    """Represents a summarized model or theory."""
    name: str
    assumptions: List[str] = field(default_factory=list)
    objective: str = ""
    formulation: str = ""
    training: str = ""
    properties: List[str] = field(default_factory=list)
    practical_notes: List[str] = field(default_factory=list)
    full_summary: str = ""
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "assumptions": self.assumptions,
            "objective": self.objective,
            "formulation": self.formulation,
            "training": self.training,
            "properties": self.properties,
            "practical_notes": self.practical_notes,
            "full_summary": self.full_summary
        }


class ModelSummarizer:
    """
    Chain for summarizing models and theoretical frameworks.
    
    Extracts key components:
    - Assumptions
    - Objectives/Loss functions
    - Model formulation
    - Training/Optimization procedures
    - Key properties
    """
    
    def __init__(
        self,
        retrieval_service: RetrievalService,
        llm_client: LLMClient,
        include_assumptions: bool = True,
        include_objectives: bool = True,
        include_training: bool = True
    ):
        """
        Initialize model summarizer.
        
        Args:
            retrieval_service: Service for retrieving context
            llm_client: LLM for generation
            include_assumptions: Include model assumptions
            include_objectives: Include objectives/loss functions
            include_training: Include training procedures
        """
        self.retrieval_service = retrieval_service
        self.llm_client = llm_client
        self.include_assumptions = include_assumptions
        self.include_objectives = include_objectives
        self.include_training = include_training
    
    @trace_function(name="summarize_model")
    def summarize(
        self,
        model_name: str,
        queries: Optional[List[str]] = None,
        config: Optional[GenerationConfig] = None
    ) -> ModelSummary:
        """
        Summarize a model or theoretical framework.
        
        Args:
            model_name: Name of the model to summarize
            queries: Optional specific queries
            config: Generation configuration
            
        Returns:
            ModelSummary
        """
        config = config or GenerationConfig()
        
        # Build queries based on model name
        if queries is None:
            queries = [
                f"{model_name} model description",
                f"{model_name} assumptions",
                f"{model_name} objective loss function",
                f"{model_name} training optimization algorithm",
                f"{model_name} properties theoretical results",
            ]
        
        # Retrieve relevant content
        all_results = []
        for query in queries:
            results = self.retrieval_service.retrieve(query, top_k=3)
            all_results.extend(results)
        
        # Deduplicate results
        seen = set()
        unique_results = []
        for r in all_results:
            key = r.content[:200]
            if key not in seen:
                unique_results.append(r)
                seen.add(key)
        
        if not unique_results:
            logger.warning(f"No content found for model: {model_name}")
            return ModelSummary(name=model_name)
        
        # Generate summary
        summary = self._generate_summary(model_name, unique_results, config)
        
        logger.info(f"Summarized model: {model_name}")
        return summary
    
    def _generate_summary(
        self,
        model_name: str,
        results: List[RetrievalResult],
        config: GenerationConfig
    ) -> ModelSummary:
        """Generate model summary from retrieval results."""
        context = "\n\n".join(r.content for r in results)
        
        prompt = get_prompt("model_extraction")(
            context=context,
            model_name=model_name
        )
        
        result = self.llm_client.generate(prompt, config)
        
        # Parse the structured response
        return self._parse_response(model_name, result.text)
    
    def _parse_response(self, model_name: str, text: str) -> ModelSummary:
        """Parse model summarization response."""
        summary = ModelSummary(name=model_name, full_summary=text)
        
        current_section = None
        current_content = []
        
        for line in text.split("\n"):
            line_stripped = line.strip()
            
            # Detect section headers
            if line_stripped.startswith("### Assumptions"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "assumptions"
                current_content = []
            elif line_stripped.startswith("### Objective"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "objective"
                current_content = []
            elif line_stripped.startswith("### Formulation"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "formulation"
                current_content = []
            elif line_stripped.startswith("### Training") or line_stripped.startswith("### Optimization"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "training"
                current_content = []
            elif line_stripped.startswith("### Key Properties") or line_stripped.startswith("### Properties"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "properties"
                current_content = []
            elif line_stripped.startswith("### Practical"):
                if current_section and current_content:
                    self._assign_section(summary, current_section, current_content)
                current_section = "practical"
                current_content = []
            elif line_stripped and not line_stripped.startswith("##"):
                current_content.append(line_stripped)
        
        # Assign final section
        if current_section and current_content:
            self._assign_section(summary, current_section, current_content)
        
        return summary
    
    def _assign_section(
        self,
        summary: ModelSummary,
        section: str,
        content: List[str]
    ) -> None:
        """Assign content to the appropriate section."""
        if section == "assumptions" and self.include_assumptions:
            summary.assumptions = self._parse_list(content)
        elif section == "objective" and self.include_objectives:
            summary.objective = "\n".join(content)
        elif section == "formulation":
            summary.formulation = "\n".join(content)
        elif section == "training" and self.include_training:
            summary.training = "\n".join(content)
        elif section == "properties":
            summary.properties = self._parse_list(content)
        elif section == "practical":
            summary.practical_notes = self._parse_list(content)
    
    def _parse_list(self, content: List[str]) -> List[str]:
        """Parse content as a list of items."""
        items = []
        current_item = []
        
        for line in content:
            if line.startswith("- ") or line.startswith("* "):
                if current_item:
                    items.append(" ".join(current_item))
                current_item = [line[2:]]
            elif line.startswith(tuple(f"{i}." for i in range(1, 10))):
                if current_item:
                    items.append(" ".join(current_item))
                current_item = [line.split(".", 1)[1].strip()]
            elif current_item:
                current_item.append(line)
        
        if current_item:
            items.append(" ".join(current_item))
        
        return items if items else content
    
    @trace_function(name="extract_all_models")
    def extract_all(
        self,
        config: Optional[GenerationConfig] = None
    ) -> List[ModelSummary]:
        """
        Extract all models/theories from the document.
        
        Args:
            config: Generation configuration
            
        Returns:
            List of ModelSummary
        """
        config = config or GenerationConfig()
        
        # First, identify model names
        results = self.retrieval_service.retrieve(
            "model architecture method approach algorithm",
            top_k=10
        )
        
        # Extract potential model names from content
        model_names = self._identify_models(results)
        
        if not model_names:
            # Use generic approach
            model_names = ["Main Model"]
        
        # Summarize each model
        summaries = []
        for name in model_names[:5]:  # Limit to 5 models
            summary = self.summarize(name, config=config)
            if summary.objective or summary.formulation:
                summaries.append(summary)
        
        return summaries
    
    def _identify_models(self, results: List[RetrievalResult]) -> List[str]:
        """Identify model names from retrieval results."""
        import re
        
        model_names = set()
        
        # Patterns for model names
        patterns = [
            r'(\w+(?:\s+\w+)?)\s+(?:model|algorithm|method|approach)',
            r'(?:the|our)\s+(\w+(?:\s+\w+)?)\s+(?:model|algorithm)',
            r'(\w+(?:-\w+)?)\s+(?:network|architecture)',
        ]
        
        for result in results:
            text = result.content
            for pattern in patterns:
                for match in re.finditer(pattern, text, re.IGNORECASE):
                    name = match.group(1).strip()
                    if len(name) > 2 and name.lower() not in ["the", "our", "this", "that"]:
                        model_names.add(name)
        
        return list(model_names)[:10]
    
    def format_for_latex(self, summary: ModelSummary) -> str:
        """
        Format model summary for LaTeX.
        
        Args:
            summary: ModelSummary to format
            
        Returns:
            LaTeX formatted string
        """
        lines = [f"\\subsection{{{summary.name}}}"]
        
        if summary.assumptions:
            lines.append("\\paragraph{Assumptions}")
            lines.append("\\begin{itemize}")
            for assumption in summary.assumptions:
                lines.append(f"  \\item {assumption}")
            lines.append("\\end{itemize}")
        
        if summary.objective:
            lines.append("\\paragraph{Objective}")
            lines.append(summary.objective)
        
        if summary.formulation:
            lines.append("\\paragraph{Formulation}")
            lines.append(summary.formulation)
        
        if summary.training:
            lines.append("\\paragraph{Training}")
            lines.append(summary.training)
        
        if summary.properties:
            lines.append("\\paragraph{Key Properties}")
            lines.append("\\begin{itemize}")
            for prop in summary.properties:
                lines.append(f"  \\item {prop}")
            lines.append("\\end{itemize}")
        
        return "\n".join(lines)



"""Infrastructure layer - adapters for vector DB, model backends, config, logging, tracing."""

from .settings import Settings
from .vectordb import VectorStore, ChromaStore, QdrantStore
from .logging import get_logger, setup_logging
from .tracing import Tracer, trace_function

__all__ = [
    "Settings",
    "VectorStore",
    "ChromaStore", 
    "QdrantStore",
    "get_logger",
    "setup_logging",
    "Tracer",
    "trace_function",
]



"""
LaTeX document builder for cheat sheet generation.

Creates structured LaTeX documents from CheatSheetContent
and compiles them to PDF.
"""

import os
import re
import subprocess
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from jinja2 import Environment, BaseLoader

from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


# Default LaTeX template for cheat sheets
DEFAULT_TEMPLATE = r"""
\documentclass[{{ font_size }}pt,{{ paper_size }}]{article}

% Packages
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{geometry}
\usepackage{multicol}
\usepackage{amsmath,amssymb,amsthm}
\usepackage{algorithm}
\usepackage{algpseudocode}
\usepackage{listings}
\usepackage{xcolor}
\usepackage{hyperref}
\usepackage{titlesec}
\usepackage{enumitem}
\usepackage{fancyhdr}
\usepackage{parskip}

% Page geometry
\geometry{
    top={{ margins.top }},
    bottom={{ margins.bottom }},
    left={{ margins.left }},
    right={{ margins.right }}
}

% Colors
\definecolor{sectioncolor}{RGB}{0,51,102}
\definecolor{codebackground}{RGB}{248,248,248}
\definecolor{codecomment}{RGB}{106,153,85}
\definecolor{codekeyword}{RGB}{0,0,255}

% Section formatting
\titleformat{\section}
    {\normalfont\large\bfseries\color{sectioncolor}}
    {\thesection}{1em}{}
\titleformat{\subsection}
    {\normalfont\normalsize\bfseries\color{sectioncolor}}
    {\thesubsection}{1em}{}
\titlespacing*{\section}{0pt}{2ex plus 1ex minus .2ex}{1ex plus .2ex}
\titlespacing*{\subsection}{0pt}{1.5ex plus 1ex minus .2ex}{0.5ex plus .2ex}

% List formatting
\setlist[itemize]{noitemsep, topsep=0pt, leftmargin=*}
\setlist[enumerate]{noitemsep, topsep=0pt, leftmargin=*}

% Code listing style
\lstset{
    basicstyle=\ttfamily\small,
    backgroundcolor=\color{codebackground},
    commentstyle=\color{codecomment},
    keywordstyle=\color{codekeyword},
    breaklines=true,
    frame=single,
    framesep=2pt,
    xleftmargin=5pt,
    xrightmargin=5pt
}

% Header/Footer
\pagestyle{fancy}
\fancyhf{}
\fancyhead[L]{\small {{ title }}}
\fancyhead[R]{\small \thepage}
\renewcommand{\headrulewidth}{0.4pt}

% Theorem environments
\newtheorem{theorem}{Theorem}
\newtheorem{lemma}{Lemma}
\newtheorem{definition}{Definition}

% Custom commands
\newcommand{\important}[1]{\textbf{#1}}
\newcommand{\concept}[1]{\textit{#1}}

\begin{document}

{% if include_toc %}
\tableofcontents
\newpage
{% endif %}

% Title
\begin{center}
    {\LARGE\bfseries {{ title }}}\\[0.5em]
    {% if authors %}
    {\normalsize {{ authors | join(', ') }}}\\[0.3em]
    {% endif %}
    {\small\today}
\end{center}

\vspace{1em}

{% if columns > 1 %}
\begin{multicols}{ {{ columns }} }
{% endif %}

% Overview
{% if overview %}
\section*{Overview}
{{ overview }}
{% endif %}

% Core Concepts
{% if concepts %}
\section{Core Concepts}
{% for concept in concepts %}
\textbf{ {{ concept.name }} }{% if concept.definition %}: {{ concept.definition }}{% endif %}
{% if concept.intuition %}
\textit{Intuition:} {{ concept.intuition }}
{% endif %}
{% endfor %}
{% endif %}

% Key Formulas
{% if formulas %}
\section{Key Formulas}
{% for group_name, group_formulas in formulas_grouped.items() %}
\subsection{ {{ group_name }} }
{% for formula in group_formulas %}
{% if formula.name %}\textbf{ {{ formula.name }} }:{% endif %}
{% if formula.latex %}
{{ formula.latex }}
{% endif %}
{% if formula.context %}
\textit{ {{ formula.context }} }
{% endif %}

{% endfor %}
{% endfor %}
{% endif %}

% Models
{% if models %}
\section{Models \& Theory}
{% for model in models %}
{% if model.summary %}
{{ model.summary }}
{% else %}
\subsection{ {{ model.name }} }
{% if model.assumptions %}
\textbf{Assumptions:}
\begin{itemize}
{% for assumption in model.assumptions %}
\item {{ assumption }}
{% endfor %}
\end{itemize}
{% endif %}
{% if model.objective %}
\textbf{Objective:} {{ model.objective }}
{% endif %}
{% if model.formulation %}
\textbf{Formulation:} {{ model.formulation }}
{% endif %}
{% endif %}
{% endfor %}
{% endif %}

% Algorithms
{% if algorithms %}
\section{Algorithms}
{% for algo in algorithms %}
\subsection{ {{ algo.name }} }
{% if algo.purpose %}
\textbf{Purpose:} {{ algo.purpose }}
{% endif %}

{% if algo.input %}
\textbf{Input:} {{ algo.input }}
{% endif %}

{% if algo.output %}
\textbf{Output:} {{ algo.output }}
{% endif %}

{% if algo.pseudocode %}
\begin{lstlisting}
{{ algo.pseudocode }}
\end{lstlisting}
{% endif %}

{% if algo.time_complexity or algo.space_complexity %}
\textbf{Complexity:} {% if algo.time_complexity %}Time: {{ algo.time_complexity }}{% endif %}{% if algo.space_complexity %}, Space: {{ algo.space_complexity }}{% endif %}
{% endif %}

{% if algo.key_insight %}
\textbf{Key Insight:} {{ algo.key_insight }}
{% endif %}

{% endfor %}
{% endif %}

% Connections
{% if connections %}
\section{Related Topics}
\begin{itemize}
{% for conn in connections %}
\item \textbf{ {{ conn.topic }} }: {{ conn.description }}
{% endfor %}
\end{itemize}
{% endif %}

% Recap
{% if recap %}
\section{Quick Reference}
{{ recap }}
{% endif %}

{% if columns > 1 %}
\end{multicols}
{% endif %}

% References
{% if include_references and references %}
\section*{References}
\begin{enumerate}
{% for ref in references %}
\item {{ ref }}
{% endfor %}
\end{enumerate}
{% endif %}

\end{document}
"""


@dataclass
class CheatSheet:
    """Represents a complete cheat sheet document."""
    title: str
    content: Dict[str, Any]
    latex_source: str = ""
    pdf_path: Optional[Path] = None
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "title": self.title,
            "content": self.content,
            "has_pdf": self.pdf_path is not None
        }


class LaTeXBuilder:
    """
    Builder for creating LaTeX cheat sheets.
    
    Uses Jinja2 templates for flexible document generation.
    """
    
    def __init__(
        self,
        settings=None,
        template: Optional[str] = None
    ):
        """
        Initialize LaTeX builder.
        
        Args:
            settings: LaTeXSettings or dict with settings
            template: Custom Jinja2 template string
        """
        self.settings = settings
        self.template_str = template or DEFAULT_TEMPLATE
        
        # Initialize Jinja2 environment
        self.env = Environment(loader=BaseLoader())
        self.env.filters['escape_latex'] = self._escape_latex
        
        # Compile template
        self.template = self.env.from_string(self.template_str)
    
    @staticmethod
    def _escape_latex(text: str) -> str:
        """Escape special LaTeX characters."""
        if not text:
            return ""
        
        # Characters that need escaping
        chars = {
            '&': r'\&',
            '%': r'\%',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '~': r'\textasciitilde{}',
            '^': r'\textasciicircum{}',
        }
        
        # Don't escape if already in math mode or LaTeX command
        if text.startswith('$') or text.startswith('\\'):
            return text
        
        for char, replacement in chars.items():
            text = text.replace(char, replacement)
        
        return text
    
    @trace_function(name="build_latex")
    def build(self, content: "CheatSheetContent") -> str:
        """
        Build LaTeX source from content.
        
        Args:
            content: CheatSheetContent to render
            
        Returns:
            LaTeX source string
        """
        # Prepare template context
        context = self._prepare_context(content)
        
        # Render template
        latex_source = self.template.render(**context)
        
        # Post-process
        latex_source = self._post_process(latex_source)
        
        logger.info(f"Built LaTeX document: {len(latex_source)} characters")
        return latex_source
    
    def _prepare_context(self, content: "CheatSheetContent") -> Dict[str, Any]:
        """Prepare template context from content and settings."""
        # Get settings or defaults
        if self.settings:
            font_size = self.settings.font_size
            paper_size = self.settings.paper_size
            margins = {
                "top": self.settings.margins.top,
                "bottom": self.settings.margins.bottom,
                "left": self.settings.margins.left,
                "right": self.settings.margins.right,
            }
            columns = self.settings.columns
            include_toc = self.settings.include_toc
            include_references = self.settings.include_references
        else:
            font_size = 10
            paper_size = "a4paper"
            margins = {"top": "1.5cm", "bottom": "1.5cm", "left": "1.5cm", "right": "1.5cm"}
            columns = 2
            include_toc = False
            include_references = True
        
        # Group formulas by topic
        formulas_grouped = self._group_formulas(content.formulas)
        
        return {
            "title": content.title,
            "authors": content.references[:1] if content.references else [],
            "overview": content.overview,
            "concepts": content.concepts,
            "formulas": content.formulas,
            "formulas_grouped": formulas_grouped,
            "models": content.models,
            "algorithms": content.algorithms,
            "connections": content.connections,
            "recap": content.recap,
            "references": content.references,
            "font_size": font_size,
            "paper_size": paper_size,
            "margins": margins,
            "columns": columns,
            "include_toc": include_toc,
            "include_references": include_references,
        }
    
    def _group_formulas(self, formulas: List[Dict[str, Any]]) -> Dict[str, List[Dict[str, Any]]]:
        """Group formulas by topic."""
        groups: Dict[str, List[Dict[str, Any]]] = {}
        
        for formula in formulas:
            topic = formula.get("topic", "General")
            if not topic:
                topic = "General"
            
            if topic not in groups:
                groups[topic] = []
            groups[topic].append(formula)
        
        return groups if groups else {"General": formulas}
    
    def _post_process(self, latex: str) -> str:
        """Post-process LaTeX source."""
        # Remove excessive blank lines
        latex = re.sub(r'\n{3,}', '\n\n', latex)
        
        # Fix common issues
        latex = re.sub(r'\\textbf\{\s*\}', '', latex)  # Remove empty bold
        latex = re.sub(r'\\textit\{\s*\}', '', latex)  # Remove empty italic
        
        return latex.strip()
    
    @trace_function(name="compile_pdf")
    def compile_pdf(
        self,
        latex_source: str,
        output_path: Union[str, Path],
        compiler: Optional[str] = None
    ) -> Path:
        """
        Compile LaTeX source to PDF.
        
        Args:
            latex_source: LaTeX source string
            output_path: Path for output PDF
            compiler: LaTeX compiler to use (pdflatex, xelatex, lualatex, tectonic)
            
        Returns:
            Path to generated PDF
        """
        output_path = Path(output_path)
        
        # Get compiler from settings or default
        if compiler is None:
            compiler = self.settings.compiler if self.settings else "pdflatex"
        
        # Create temporary directory for compilation
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            tex_file = tmpdir / "cheatsheet.tex"
            
            # Write LaTeX source
            tex_file.write_text(latex_source, encoding='utf-8')
            
            # Compile
            try:
                if compiler == "tectonic":
                    self._compile_with_tectonic(tex_file, tmpdir)
                else:
                    self._compile_with_latex(tex_file, tmpdir, compiler)
                
                # Copy PDF to output location
                pdf_file = tmpdir / "cheatsheet.pdf"
                if pdf_file.exists():
                    output_path.parent.mkdir(parents=True, exist_ok=True)
                    
                    import shutil
                    shutil.copy(pdf_file, output_path)
                    
                    logger.info(f"Generated PDF: {output_path}")
                    return output_path
                else:
                    raise RuntimeError("PDF not generated")
                    
            except subprocess.CalledProcessError as e:
                # Try to get error from log file
                log_file = tmpdir / "cheatsheet.log"
                error_msg = ""
                if log_file.exists():
                    log_content = log_file.read_text(encoding='utf-8', errors='ignore')
                    # Extract error lines
                    for line in log_content.split('\n'):
                        if '!' in line or 'Error' in line:
                            error_msg += line + '\n'
                
                logger.error(f"LaTeX compilation failed: {error_msg or str(e)}")
                raise RuntimeError(f"LaTeX compilation failed: {error_msg or str(e)}")
    
    def _compile_with_latex(
        self,
        tex_file: Path,
        tmpdir: Path,
        compiler: str
    ) -> None:
        """Compile with standard LaTeX compiler."""
        # Run twice for references
        for _ in range(2):
            result = subprocess.run(
                [
                    compiler,
                    "-interaction=nonstopmode",
                    "-halt-on-error",
                    tex_file.name
                ],
                cwd=tmpdir,
                capture_output=True,
                text=True,
                timeout=120
            )
            
            if result.returncode != 0:
                # Check if it's a real error or just warnings
                log_file = tmpdir / "cheatsheet.log"
                if log_file.exists():
                    log_content = log_file.read_text(encoding='utf-8', errors='ignore')
                    if "Fatal error" in log_content or "! " in log_content:
                        raise subprocess.CalledProcessError(
                            result.returncode, compiler, result.stdout, result.stderr
                        )
    
    def _compile_with_tectonic(self, tex_file: Path, tmpdir: Path) -> None:
        """Compile with tectonic."""
        result = subprocess.run(
            ["tectonic", tex_file.name],
            cwd=tmpdir,
            capture_output=True,
            text=True,
            timeout=120
        )
        
        if result.returncode != 0:
            raise subprocess.CalledProcessError(
                result.returncode, "tectonic", result.stdout, result.stderr
            )
    
    def save_latex(self, latex_source: str, output_path: Union[str, Path]) -> Path:
        """
        Save LaTeX source to file.
        
        Args:
            latex_source: LaTeX source string
            output_path: Output file path
            
        Returns:
            Path to saved file
        """
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        output_path.write_text(latex_source, encoding='utf-8')
        logger.info(f"Saved LaTeX source: {output_path}")
        
        return output_path
    
    def build_and_compile(
        self,
        content: "CheatSheetContent",
        output_path: Union[str, Path]
    ) -> CheatSheet:
        """
        Build LaTeX and compile to PDF.
        
        Args:
            content: CheatSheetContent
            output_path: Output PDF path
            
        Returns:
            CheatSheet with source and PDF path
        """
        # Build LaTeX
        latex_source = self.build(content)
        
        # Compile to PDF
        pdf_path = self.compile_pdf(latex_source, output_path)
        
        return CheatSheet(
            title=content.title,
            content=content.to_dict(),
            latex_source=latex_source,
            pdf_path=pdf_path
        )


def check_latex_installation() -> Dict[str, bool]:
    """
    Check which LaTeX compilers are available.
    
    Returns:
        Dictionary of compiler name to availability
    """
    compilers = ["pdflatex", "xelatex", "lualatex", "tectonic"]
    available = {}
    
    for compiler in compilers:
        try:
            result = subprocess.run(
                [compiler, "--version"],
                capture_output=True,
                timeout=5
            )
            available[compiler] = result.returncode == 0
        except (subprocess.TimeoutExpired, FileNotFoundError):
            available[compiler] = False
    
    return available



"""
FastAPI application for ReadAnyBook REST API.

Provides HTTP endpoints for cheat sheet generation.
"""

import asyncio
import tempfile
import uuid
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Dict, List, Optional

try:
    from fastapi import FastAPI, HTTPException, BackgroundTasks, UploadFile, File, Query
    from fastapi.responses import FileResponse, JSONResponse
    from fastapi.middleware.cors import CORSMiddleware
    from pydantic import BaseModel, Field
    FASTAPI_AVAILABLE = True
except ImportError:
    FASTAPI_AVAILABLE = False


# Pydantic models for API
if FASTAPI_AVAILABLE:
    
    class GenerationRequest(BaseModel):
        """Request to generate a cheat sheet."""
        title: Optional[str] = Field(None, description="Title for the cheat sheet")
        collection_name: Optional[str] = Field("default", description="Vector store collection")
        profile: Optional[str] = Field(None, description="Configuration profile")
        include_concepts: bool = Field(True, description="Include concepts section")
        include_formulas: bool = Field(True, description="Include formulas section")
        include_algorithms: bool = Field(True, description="Include algorithms section")
        include_models: bool = Field(True, description="Include models section")
        latex_only: bool = Field(False, description="Return LaTeX only, don't compile")
    
    
    class SearchRequest(BaseModel):
        """Request to search indexed documents."""
        query: str = Field(..., description="Search query")
        collection_name: str = Field("default", description="Collection to search")
        top_k: int = Field(5, ge=1, le=50, description="Number of results")
        mode: str = Field("hybrid", description="Search mode: dense, sparse, hybrid")
    
    
    class SearchResult(BaseModel):
        """Search result."""
        content: str
        score: float
        metadata: Dict[str, Any] = {}
    
    
    class IndexRequest(BaseModel):
        """Request to index a document."""
        collection_name: str = Field("default", description="Collection name")
    
    
    class TaskStatus(BaseModel):
        """Background task status."""
        task_id: str
        status: str  # pending, processing, completed, failed
        progress: float = 0.0
        result: Optional[Dict[str, Any]] = None
        error: Optional[str] = None
    
    
    class CheatSheetContent(BaseModel):
        """Generated cheat sheet content."""
        title: str
        overview: Optional[str] = None
        concepts: List[Dict[str, Any]] = []
        formulas: List[Dict[str, Any]] = []
        algorithms: List[Dict[str, Any]] = []
        models: List[Dict[str, Any]] = []
    
    
    # Store for background tasks
    task_store: Dict[str, TaskStatus] = {}
    
    # Pipeline instance (singleton)
    _pipeline = None
    
    
    @asynccontextmanager
    async def lifespan(app: FastAPI):
        """Application lifespan handler."""
        # Startup
        from ..infra.logging import setup_logging
        setup_logging(level="INFO", pretty=True)
        
        yield
        
        # Shutdown
        global _pipeline
        _pipeline = None
    
    
    def create_app(settings=None) -> FastAPI:
        """
        Create FastAPI application.
        
        Args:
            settings: Optional Settings instance
            
        Returns:
            FastAPI application
        """
        app = FastAPI(
            title="ReadAnyBook API",
            description="Generate cheat sheets from books and papers using RAG",
            version="0.1.0",
            lifespan=lifespan,
        )
        
        # Add CORS middleware
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        
        # Store settings in app state
        app.state.settings = settings
        
        # Register routes
        register_routes(app)
        
        return app
    
    
    def get_pipeline():
        """Get or create pipeline instance."""
        global _pipeline
        
        if _pipeline is None:
            from ..core.pipeline import CheatSheetPipeline
            from ..infra.settings import Settings
            
            settings = Settings()
            _pipeline = CheatSheetPipeline(settings)
        
        return _pipeline
    
    
    def register_routes(app: FastAPI):
        """Register API routes."""
        
        @app.get("/")
        async def root():
            """Root endpoint."""
            return {
                "name": "ReadAnyBook API",
                "version": "0.1.0",
                "docs": "/docs",
            }
        
        @app.get("/health")
        async def health():
            """Health check endpoint."""
            return {"status": "healthy"}
        
        @app.post("/upload")
        async def upload_document(
            file: UploadFile = File(...),
            collection_name: str = Query("default"),
            background_tasks: BackgroundTasks = None,
        ):
            """
            Upload and index a document.
            
            Accepts PDF, EPUB, HTML, Markdown files.
            """
            # Validate file type
            allowed_extensions = {".pdf", ".epub", ".html", ".md", ".tex"}
            file_ext = Path(file.filename).suffix.lower()
            
            if file_ext not in allowed_extensions:
                raise HTTPException(
                    status_code=400,
                    detail=f"Unsupported file type: {file_ext}. Allowed: {allowed_extensions}"
                )
            
            # Create task ID
            task_id = str(uuid.uuid4())
            task_store[task_id] = TaskStatus(
                task_id=task_id,
                status="pending",
            )
            
            # Save uploaded file
            with tempfile.NamedTemporaryFile(delete=False, suffix=file_ext) as tmp:
                content = await file.read()
                tmp.write(content)
                tmp_path = tmp.name
            
            # Process in background
            if background_tasks:
                background_tasks.add_task(
                    process_upload,
                    task_id,
                    tmp_path,
                    collection_name,
                )
            else:
                # Synchronous processing
                await asyncio.to_thread(
                    process_upload_sync,
                    task_id,
                    tmp_path,
                    collection_name,
                )
            
            return {
                "task_id": task_id,
                "status": "pending",
                "message": "Document uploaded and queued for processing",
            }
        
        @app.get("/task/{task_id}")
        async def get_task_status(task_id: str):
            """Get status of a background task."""
            if task_id not in task_store:
                raise HTTPException(status_code=404, detail="Task not found")
            
            return task_store[task_id]
        
        @app.post("/search")
        async def search(request: SearchRequest):
            """
            Search indexed documents.
            """
            try:
                pipeline = get_pipeline()
                
                # Ensure index is connected
                if not pipeline._retriever:
                    raise HTTPException(
                        status_code=400,
                        detail="No index loaded. Upload a document first."
                    )
                
                # Perform search
                results = pipeline._retriever.retrieve(
                    request.query,
                    top_k=request.top_k,
                    collection_name=request.collection_name,
                )
                
                return {
                    "query": request.query,
                    "results": [
                        SearchResult(
                            content=r.get("content", ""),
                            score=r.get("score", 0.0),
                            metadata=r.get("metadata", {}),
                        )
                        for r in results
                    ],
                }
            
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))
        
        @app.post("/generate")
        async def generate_cheatsheet(
            request: GenerationRequest,
            background_tasks: BackgroundTasks,
        ):
            """
            Generate a cheat sheet from indexed documents.
            
            Returns task ID for tracking progress.
            """
            task_id = str(uuid.uuid4())
            task_store[task_id] = TaskStatus(
                task_id=task_id,
                status="pending",
            )
            
            background_tasks.add_task(
                process_generation,
                task_id,
                request,
            )
            
            return {
                "task_id": task_id,
                "status": "pending",
                "message": "Generation started",
            }
        
        @app.get("/download/{task_id}")
        async def download_result(task_id: str):
            """
            Download generated cheat sheet.
            """
            if task_id not in task_store:
                raise HTTPException(status_code=404, detail="Task not found")
            
            task = task_store[task_id]
            
            if task.status != "completed":
                raise HTTPException(
                    status_code=400,
                    detail=f"Task not completed. Status: {task.status}"
                )
            
            if not task.result or "file_path" not in task.result:
                raise HTTPException(status_code=404, detail="No file generated")
            
            file_path = Path(task.result["file_path"])
            if not file_path.exists():
                raise HTTPException(status_code=404, detail="File not found")
            
            return FileResponse(
                file_path,
                media_type="application/pdf" if file_path.suffix == ".pdf" else "text/plain",
                filename=file_path.name,
            )
        
        @app.get("/collections")
        async def list_collections():
            """List available collections."""
            try:
                pipeline = get_pipeline()
                
                if pipeline._vector_store:
                    collections = pipeline._vector_store.list_collections()
                    return {"collections": collections}
                
                return {"collections": []}
            
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))
        
        @app.delete("/collections/{name}")
        async def delete_collection(name: str):
            """Delete a collection."""
            try:
                pipeline = get_pipeline()
                
                if pipeline._vector_store:
                    pipeline._vector_store.delete_collection(name)
                    return {"message": f"Collection '{name}' deleted"}
                
                raise HTTPException(status_code=400, detail="No vector store configured")
            
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))
    
    
    def process_upload_sync(task_id: str, file_path: str, collection_name: str):
        """Synchronously process uploaded document."""
        try:
            task_store[task_id].status = "processing"
            task_store[task_id].progress = 0.1
            
            pipeline = get_pipeline()
            
            # Ingest
            task_store[task_id].progress = 0.3
            pipeline.ingest(file_path)
            
            # Index
            task_store[task_id].progress = 0.6
            num_chunks = pipeline.index(collection_name=collection_name)
            
            task_store[task_id].status = "completed"
            task_store[task_id].progress = 1.0
            task_store[task_id].result = {
                "chunks_indexed": num_chunks,
                "collection": collection_name,
            }
            
        except Exception as e:
            task_store[task_id].status = "failed"
            task_store[task_id].error = str(e)
        
        finally:
            # Clean up temp file
            Path(file_path).unlink(missing_ok=True)
    
    
    async def process_upload(task_id: str, file_path: str, collection_name: str):
        """Process uploaded document in background."""
        await asyncio.to_thread(process_upload_sync, task_id, file_path, collection_name)
    
    
    async def process_generation(task_id: str, request: GenerationRequest):
        """Process cheat sheet generation in background."""
        try:
            task_store[task_id].status = "processing"
            task_store[task_id].progress = 0.1
            
            pipeline = get_pipeline()
            
            # Generate content
            task_store[task_id].progress = 0.3
            content = await asyncio.to_thread(
                pipeline.generate_content,
                request.collection_name,
            )
            
            # Set title
            if request.title:
                content.title = request.title
            
            task_store[task_id].progress = 0.7
            
            # Build output
            with tempfile.NamedTemporaryFile(
                delete=False,
                suffix=".tex" if request.latex_only else ".pdf"
            ) as tmp:
                output_path = tmp.name
            
            if request.latex_only:
                latex = pipeline.build_latex(content)
                Path(output_path).write_text(latex, encoding='utf-8')
            else:
                await asyncio.to_thread(pipeline.build, content, output_path)
            
            task_store[task_id].status = "completed"
            task_store[task_id].progress = 1.0
            task_store[task_id].result = {
                "file_path": output_path,
                "title": content.title,
                "concepts_count": len(content.concepts),
                "formulas_count": len(content.formulas),
                "algorithms_count": len(content.algorithms),
                "models_count": len(content.models),
            }
            
        except Exception as e:
            task_store[task_id].status = "failed"
            task_store[task_id].error = str(e)
    
    
    # Create default app instance
    app = create_app()

else:
    # Fallback when FastAPI not installed
    def create_app(settings=None):
        raise ImportError("FastAPI not installed. Run: pip install fastapi uvicorn")
    
    app = None



"""
Prompt templates for different generation tasks.

Each prompt is designed for specific extraction/generation tasks
in the cheat sheet pipeline.
"""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from string import Template


@dataclass
class PromptTemplate:
    """Template for generating prompts."""
    template: str
    input_variables: List[str]
    description: str = ""
    
    def format(self, **kwargs) -> str:
        """Format the template with provided variables."""
        # Validate all required variables are provided
        missing = set(self.input_variables) - set(kwargs.keys())
        if missing:
            raise ValueError(f"Missing required variables: {missing}")
        
        return Template(self.template).safe_substitute(**kwargs)
    
    def __call__(self, **kwargs) -> str:
        """Allow calling template directly."""
        return self.format(**kwargs)


# =============================================================================
# Concept Extraction Prompts
# =============================================================================

CONCEPT_EXTRACTION_PROMPT = PromptTemplate(
    template="""You are an expert at extracting and summarizing key concepts from technical documents.

Given the following text from a book or paper, extract the main concepts with clear, concise definitions and intuitions.

TEXT:
$context

INSTRUCTIONS:
1. Identify the most important concepts (definitions, terms, ideas)
2. For each concept, provide:
   - Name: The concept name
   - Definition: A precise, formal definition
   - Intuition: A simple, intuitive explanation
   - Related concepts: Other concepts it connects to
3. Focus on concepts that are essential for understanding the material
4. Use LaTeX notation for any mathematical symbols: $symbol$

Format your response as a structured list:

CONCEPT: [Name]
DEFINITION: [Formal definition]
INTUITION: [Simple explanation]
RELATED: [Comma-separated related concepts]

---

Extract the key concepts:""",
    input_variables=["context"],
    description="Extract key concepts from technical text"
)


CONCEPT_SUMMARY_PROMPT = PromptTemplate(
    template="""Summarize the following concepts into a cohesive overview for a cheat sheet.

CONCEPTS:
$concepts

INSTRUCTIONS:
1. Group related concepts together
2. Create a logical flow from basic to advanced
3. Keep definitions precise but concise
4. Use bullet points for clarity
5. Include LaTeX notation for math: $symbol$
6. Maximum length: $max_words words

Write a well-organized concept summary:""",
    input_variables=["concepts", "max_words"],
    description="Summarize extracted concepts"
)


# =============================================================================
# Formula Extraction Prompts
# =============================================================================

FORMULA_EXTRACTION_PROMPT = PromptTemplate(
    template="""You are an expert at extracting and organizing mathematical formulas from technical documents.

Given the following text, extract all important mathematical formulas and equations.

TEXT:
$context

INSTRUCTIONS:
1. Identify all significant formulas and equations
2. For each formula, provide:
   - Name/Label: What the formula represents
   - Formula: The formula in LaTeX notation
   - Variables: Define each variable
   - Context: When/why this formula is used
   - Dependencies: Other formulas it relies on
3. Normalize all formulas to consistent LaTeX syntax
4. Group formulas by topic/purpose

Format your response:

FORMULA: [Name/Label]
LATEX: $$ [LaTeX formula] $$
VARIABLES:
- [variable]: [description]
CONTEXT: [Brief explanation of usage]
DEPENDS_ON: [Related formula names]

---

Extract the formulas:""",
    input_variables=["context"],
    description="Extract mathematical formulas"
)


FORMULA_ORGANIZATION_PROMPT = PromptTemplate(
    template="""Organize the following formulas into a coherent reference section for a cheat sheet.

FORMULAS:
$formulas

TOPIC: $topic

INSTRUCTIONS:
1. Group formulas by subtopic
2. Order from fundamental to derived
3. Show relationships between formulas
4. Keep LaTeX notation clean and consistent
5. Add brief explanatory notes where helpful
6. Maximum $max_formulas formulas

Create an organized formula reference:""",
    input_variables=["formulas", "topic", "max_formulas"],
    description="Organize formulas for cheat sheet"
)


# =============================================================================
# Model/Theory Summarization Prompts
# =============================================================================

MODEL_EXTRACTION_PROMPT = PromptTemplate(
    template="""You are an expert at summarizing machine learning and mathematical models.

Given the following text about a model or theoretical framework, create a comprehensive summary.

TEXT:
$context

MODEL/TOPIC: $model_name

INSTRUCTIONS:
1. Summarize the model's key components:
   - Assumptions: What does the model assume?
   - Objective: What is the model trying to achieve?
   - Formulation: Key equations and structure
   - Training/Optimization: How is it learned/solved?
   - Properties: Key theoretical properties
2. Use precise mathematical notation (LaTeX)
3. Be concise but complete
4. Highlight practical implications

Format your response:

## $model_name

### Assumptions
[List key assumptions]

### Objective
[Main goal and loss function/objective]

### Formulation
[Key equations with LaTeX]

### Training/Optimization
[How to train or solve]

### Key Properties
[Important theoretical results]

### Practical Notes
[Implementation tips, common pitfalls]

---

Summarize the model:""",
    input_variables=["context", "model_name"],
    description="Extract and summarize model descriptions"
)


# =============================================================================
# Algorithm Synthesis Prompts
# =============================================================================

ALGORITHM_EXTRACTION_PROMPT = PromptTemplate(
    template="""You are an expert at extracting and formalizing algorithms from technical documents.

Given the following text, extract and formalize the algorithm(s) described.

TEXT:
$context

INSTRUCTIONS:
1. Identify the main algorithm(s)
2. For each algorithm, provide:
   - Name: Algorithm name
   - Purpose: What problem it solves
   - Input: Required inputs with types
   - Output: What it produces
   - Pseudocode: Clear, numbered steps
   - Complexity: Time and space complexity
   - Key insight: The main idea that makes it work
3. Use clear pseudocode conventions
4. Note any important edge cases

Format your response:

ALGORITHM: [Name]
PURPOSE: [What it solves]
INPUT: [Input specification]
OUTPUT: [Output specification]

PSEUDOCODE:
```
1. [Step 1]
2. [Step 2]
   2.1 [Substep]
   ...
```

COMPLEXITY:
- Time: O(...)
- Space: O(...)

KEY INSIGHT: [Main algorithmic insight]
EDGE CASES: [Important edge cases to handle]

---

Extract the algorithm(s):""",
    input_variables=["context"],
    description="Extract algorithm specifications"
)


ALGORITHM_SYNTHESIS_PROMPT = PromptTemplate(
    template="""Create clean, well-documented pseudocode for the following algorithm description.

DESCRIPTION:
$description

ALGORITHM NAME: $algorithm_name

INSTRUCTIONS:
1. Write clear, numbered pseudocode
2. Use consistent indentation
3. Include comments for complex steps
4. Define any helper functions needed
5. Add complexity analysis
6. Include a brief example if helpful

Create the pseudocode:

```
Algorithm: $algorithm_name
Input: [specify]
Output: [specify]

[Pseudocode here]
```

Complexity Analysis:
- Time: 
- Space:

Example (optional):
[Brief example of execution]

---

Generate the pseudocode:""",
    input_variables=["description", "algorithm_name"],
    description="Synthesize pseudocode from description"
)


# =============================================================================
# General RAG Prompts
# =============================================================================

RAG_QA_PROMPT = PromptTemplate(
    template="""Answer the question based on the provided context. If the answer cannot be found in the context, say "I cannot find this information in the provided context."

CONTEXT:
$context

QUESTION: $question

INSTRUCTIONS:
1. Answer based only on the provided context
2. Be precise and concise
3. Use LaTeX for any math: $symbol$
4. Cite specific parts of the context if relevant
5. If uncertain, express the uncertainty

ANSWER:""",
    input_variables=["context", "question"],
    description="RAG question answering"
)


SUMMARIZATION_PROMPT = PromptTemplate(
    template="""Summarize the following text for inclusion in a technical cheat sheet.

TEXT:
$context

SECTION: $section_name
MAX LENGTH: $max_words words

INSTRUCTIONS:
1. Focus on the most important points
2. Use bullet points for clarity
3. Preserve technical accuracy
4. Include key formulas in LaTeX
5. Be concise but complete

SUMMARY:""",
    input_variables=["context", "section_name", "max_words"],
    description="Summarization for cheat sheet"
)


# =============================================================================
# Cheat Sheet Assembly Prompts
# =============================================================================

OVERVIEW_PROMPT = PromptTemplate(
    template="""Create a brief overview and roadmap for a cheat sheet about the following topic.

TOPIC: $topic
SUBTOPICS: $subtopics

INSTRUCTIONS:
1. Write a 2-3 sentence overview
2. Explain what the reader will learn
3. List the main sections with one-line descriptions
4. Keep it concise and engaging

OVERVIEW:""",
    input_variables=["topic", "subtopics"],
    description="Generate cheat sheet overview"
)


RECAP_PROMPT = PromptTemplate(
    template="""Create a "If you remember one page..." recap section for a cheat sheet.

MAIN CONTENT:
$content

INSTRUCTIONS:
1. Distill the absolute essentials
2. Include the 3-5 most important formulas
3. List the 5-7 key concepts
4. Add 2-3 practical tips
5. This should fit on one page
6. Use bullet points

RECAP:

## If You Remember One Page...

### Core Ideas
[Most important concepts]

### Essential Formulas
[Most important equations in LaTeX]

### Quick Tips
[Practical advice]

---

Generate the recap:""",
    input_variables=["content"],
    description="Generate one-page recap"
)


CONNECTIONS_PROMPT = PromptTemplate(
    template="""Identify connections between the main topic and related areas.

MAIN TOPIC: $topic
CONTENT SUMMARY: $summary
RELATED AREAS TO CONSIDER: $related_areas

INSTRUCTIONS:
1. Identify meaningful connections
2. Explain how concepts transfer
3. Note prerequisites and follow-up topics
4. Keep entries brief

FORMAT:
- [Related topic]: [Brief description of connection]

CONNECTIONS:""",
    input_variables=["topic", "summary", "related_areas"],
    description="Generate topic connections"
)


# =============================================================================
# Prompt Registry
# =============================================================================

PROMPT_REGISTRY = {
    # Concept extraction
    "concept_extraction": CONCEPT_EXTRACTION_PROMPT,
    "concept_summary": CONCEPT_SUMMARY_PROMPT,
    
    # Formula extraction
    "formula_extraction": FORMULA_EXTRACTION_PROMPT,
    "formula_organization": FORMULA_ORGANIZATION_PROMPT,
    
    # Model summarization
    "model_extraction": MODEL_EXTRACTION_PROMPT,
    
    # Algorithm synthesis
    "algorithm_extraction": ALGORITHM_EXTRACTION_PROMPT,
    "algorithm_synthesis": ALGORITHM_SYNTHESIS_PROMPT,
    
    # General RAG
    "rag_qa": RAG_QA_PROMPT,
    "summarization": SUMMARIZATION_PROMPT,
    
    # Cheat sheet assembly
    "overview": OVERVIEW_PROMPT,
    "recap": RECAP_PROMPT,
    "connections": CONNECTIONS_PROMPT,
}


def get_prompt(name: str) -> PromptTemplate:
    """Get a prompt template by name."""
    if name not in PROMPT_REGISTRY:
        raise ValueError(f"Unknown prompt: {name}. Available: {list(PROMPT_REGISTRY.keys())}")
    return PROMPT_REGISTRY[name]


def list_prompts() -> List[str]:
    """List available prompt names."""
    return list(PROMPT_REGISTRY.keys())



"""
LLM client abstraction for model orchestration.

Supports multiple backends:
- HuggingFace Transformers
- vLLM
- Ollama
- OpenAI-compatible APIs
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from ..infra.logging import get_logger
from ..infra.tracing import trace_function

logger = get_logger(__name__)


@dataclass
class GenerationConfig:
    """Configuration for text generation."""
    max_tokens: int = 2048
    temperature: float = 0.7
    top_p: float = 0.9
    top_k: int = 50
    repetition_penalty: float = 1.1
    stop_sequences: List[str] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "repetition_penalty": self.repetition_penalty,
            "stop_sequences": self.stop_sequences
        }


@dataclass
class GenerationResult:
    """Result from LLM generation."""
    text: str
    tokens_used: int = 0
    finish_reason: str = "stop"
    model: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)


class LLMClient(ABC):
    """Abstract base class for LLM clients."""
    
    @abstractmethod
    def generate(
        self,
        prompt: str,
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """
        Generate text from prompt.
        
        Args:
            prompt: Input prompt
            config: Generation configuration
            
        Returns:
            GenerationResult
        """
        pass
    
    @abstractmethod
    def generate_batch(
        self,
        prompts: List[str],
        config: Optional[GenerationConfig] = None
    ) -> List[GenerationResult]:
        """
        Generate text for multiple prompts.
        
        Args:
            prompts: List of prompts
            config: Generation configuration
            
        Returns:
            List of GenerationResults
        """
        pass
    
    def chat(
        self,
        messages: List[Dict[str, str]],
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """
        Chat-style generation with message history.
        
        Args:
            messages: List of message dicts with 'role' and 'content'
            config: Generation configuration
            
        Returns:
            GenerationResult
        """
        # Default implementation: format messages as prompt
        prompt = self._format_chat_messages(messages)
        return self.generate(prompt, config)
    
    def _format_chat_messages(self, messages: List[Dict[str, str]]) -> str:
        """Format chat messages as a prompt."""
        formatted = []
        for msg in messages:
            role = msg.get("role", "user")
            content = msg.get("content", "")
            
            if role == "system":
                formatted.append(f"System: {content}")
            elif role == "assistant":
                formatted.append(f"Assistant: {content}")
            else:
                formatted.append(f"User: {content}")
        
        formatted.append("Assistant:")
        return "\n\n".join(formatted)


class HuggingFaceLLM(LLMClient):
    """HuggingFace Transformers LLM client."""
    
    def __init__(
        self,
        model_name: str = "mistralai/Mistral-7B-Instruct-v0.2",
        device: str = "auto",
        quantization: Optional[str] = "4bit",
        trust_remote_code: bool = True,
        torch_dtype: str = "auto"
    ):
        """
        Initialize HuggingFace LLM.
        
        Args:
            model_name: Model name or path
            device: Device to use
            quantization: Quantization mode (None, "4bit", "8bit")
            trust_remote_code: Trust remote code
            torch_dtype: Torch dtype
        """
        self.model_name = model_name
        self.device = device
        self.quantization = quantization
        self.trust_remote_code = trust_remote_code
        self.torch_dtype = torch_dtype
        
        self._model = None
        self._tokenizer = None
    
    def _load_model(self):
        """Lazy load the model."""
        if self._model is not None:
            return
        
        logger.info(f"Loading HuggingFace model: {self.model_name}")
        
        try:
            import torch
            from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig
        except ImportError:
            raise ImportError(
                "transformers and torch are required. "
                "Install with: pip install transformers torch"
            )
        
        # Resolve device
        if self.device == "auto":
            if torch.cuda.is_available():
                device = "cuda"
            elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
                device = "mps"
            else:
                device = "cpu"
        else:
            device = self.device
        
        # Resolve dtype
        if self.torch_dtype == "auto":
            dtype = torch.float16 if device != "cpu" else torch.float32
        else:
            dtype = getattr(torch, self.torch_dtype, torch.float32)
        
        # Load tokenizer
        self._tokenizer = AutoTokenizer.from_pretrained(
            self.model_name,
            trust_remote_code=self.trust_remote_code
        )
        
        # Configure quantization
        model_kwargs = {
            "trust_remote_code": self.trust_remote_code,
            "torch_dtype": dtype,
        }
        
        if self.quantization and device == "cuda":
            try:
                if self.quantization == "4bit":
                    model_kwargs["quantization_config"] = BitsAndBytesConfig(
                        load_in_4bit=True,
                        bnb_4bit_compute_dtype=dtype
                    )
                elif self.quantization == "8bit":
                    model_kwargs["quantization_config"] = BitsAndBytesConfig(
                        load_in_8bit=True
                    )
            except Exception as e:
                logger.warning(f"Quantization not available: {e}")
        else:
            model_kwargs["device_map"] = device if device != "cpu" else None
        
        # Load model
        self._model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            **model_kwargs
        )
        
        if device == "cpu":
            self._model = self._model.to(device)
        
        logger.info(f"Loaded model on {device}")
    
    @trace_function(name="hf_generate")
    def generate(
        self,
        prompt: str,
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Generate text using HuggingFace model."""
        self._load_model()
        
        config = config or GenerationConfig()
        
        # Tokenize
        inputs = self._tokenizer(prompt, return_tensors="pt")
        inputs = {k: v.to(self._model.device) for k, v in inputs.items()}
        input_length = inputs["input_ids"].shape[1]
        
        # Generate
        import torch
        
        with torch.no_grad():
            outputs = self._model.generate(
                **inputs,
                max_new_tokens=config.max_tokens,
                temperature=config.temperature,
                top_p=config.top_p,
                top_k=config.top_k,
                repetition_penalty=config.repetition_penalty,
                do_sample=config.temperature > 0,
                pad_token_id=self._tokenizer.eos_token_id
            )
        
        # Decode
        generated_tokens = outputs[0][input_length:]
        text = self._tokenizer.decode(generated_tokens, skip_special_tokens=True)
        
        return GenerationResult(
            text=text.strip(),
            tokens_used=len(generated_tokens),
            model=self.model_name
        )
    
    def generate_batch(
        self,
        prompts: List[str],
        config: Optional[GenerationConfig] = None
    ) -> List[GenerationResult]:
        """Generate text for multiple prompts."""
        return [self.generate(p, config) for p in prompts]


class OllamaLLM(LLMClient):
    """Ollama LLM client."""
    
    def __init__(
        self,
        model_name: str = "mistral",
        host: str = "http://localhost:11434"
    ):
        """
        Initialize Ollama client.
        
        Args:
            model_name: Ollama model name
            host: Ollama server URL
        """
        self.model_name = model_name
        self.host = host.rstrip("/")
    
    @trace_function(name="ollama_generate")
    def generate(
        self,
        prompt: str,
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Generate text using Ollama."""
        import requests
        
        config = config or GenerationConfig()
        
        response = requests.post(
            f"{self.host}/api/generate",
            json={
                "model": self.model_name,
                "prompt": prompt,
                "options": {
                    "num_predict": config.max_tokens,
                    "temperature": config.temperature,
                    "top_p": config.top_p,
                    "top_k": config.top_k,
                    "repeat_penalty": config.repetition_penalty
                },
                "stream": False
            }
        )
        response.raise_for_status()
        
        data = response.json()
        
        return GenerationResult(
            text=data.get("response", "").strip(),
            tokens_used=data.get("eval_count", 0),
            model=self.model_name,
            metadata={"total_duration": data.get("total_duration")}
        )
    
    def generate_batch(
        self,
        prompts: List[str],
        config: Optional[GenerationConfig] = None
    ) -> List[GenerationResult]:
        """Generate text for multiple prompts."""
        return [self.generate(p, config) for p in prompts]
    
    def chat(
        self,
        messages: List[Dict[str, str]],
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Chat using Ollama's chat API."""
        import requests
        
        config = config or GenerationConfig()
        
        response = requests.post(
            f"{self.host}/api/chat",
            json={
                "model": self.model_name,
                "messages": messages,
                "options": {
                    "num_predict": config.max_tokens,
                    "temperature": config.temperature,
                    "top_p": config.top_p
                },
                "stream": False
            }
        )
        response.raise_for_status()
        
        data = response.json()
        
        return GenerationResult(
            text=data.get("message", {}).get("content", "").strip(),
            tokens_used=data.get("eval_count", 0),
            model=self.model_name
        )


class OpenAICompatibleLLM(LLMClient):
    """OpenAI-compatible API client."""
    
    def __init__(
        self,
        model_name: str = "gpt-3.5-turbo",
        api_base: Optional[str] = None,
        api_key: Optional[str] = None
    ):
        """
        Initialize OpenAI-compatible client.
        
        Args:
            model_name: Model name
            api_base: API base URL (None for OpenAI)
            api_key: API key
        """
        self.model_name = model_name
        self.api_base = api_base
        self.api_key = api_key
        self._client = None
    
    def _get_client(self):
        """Get or create OpenAI client."""
        if self._client is not None:
            return self._client
        
        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError("openai is required. Install with: pip install openai")
        
        kwargs = {}
        if self.api_base:
            kwargs["base_url"] = self.api_base
        if self.api_key:
            kwargs["api_key"] = self.api_key
        
        self._client = OpenAI(**kwargs)
        return self._client
    
    @trace_function(name="openai_generate")
    def generate(
        self,
        prompt: str,
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Generate text using OpenAI-compatible API."""
        messages = [{"role": "user", "content": prompt}]
        return self.chat(messages, config)
    
    def generate_batch(
        self,
        prompts: List[str],
        config: Optional[GenerationConfig] = None
    ) -> List[GenerationResult]:
        """Generate text for multiple prompts."""
        return [self.generate(p, config) for p in prompts]
    
    def chat(
        self,
        messages: List[Dict[str, str]],
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Chat using OpenAI-compatible API."""
        client = self._get_client()
        config = config or GenerationConfig()
        
        response = client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            max_tokens=config.max_tokens,
            temperature=config.temperature,
            top_p=config.top_p
        )
        
        choice = response.choices[0]
        
        return GenerationResult(
            text=choice.message.content.strip(),
            tokens_used=response.usage.total_tokens if response.usage else 0,
            finish_reason=choice.finish_reason,
            model=self.model_name
        )


class VLLMClient(LLMClient):
    """vLLM client for high-throughput inference."""
    
    def __init__(
        self,
        model_name: str = "mistralai/Mistral-7B-Instruct-v0.2",
        tensor_parallel_size: int = 1
    ):
        """
        Initialize vLLM client.
        
        Args:
            model_name: Model name
            tensor_parallel_size: Number of GPUs for tensor parallelism
        """
        self.model_name = model_name
        self.tensor_parallel_size = tensor_parallel_size
        self._llm = None
    
    def _load_model(self):
        """Lazy load vLLM model."""
        if self._llm is not None:
            return
        
        try:
            from vllm import LLM
        except ImportError:
            raise ImportError("vllm is required. Install with: pip install vllm")
        
        logger.info(f"Loading vLLM model: {self.model_name}")
        
        self._llm = LLM(
            model=self.model_name,
            tensor_parallel_size=self.tensor_parallel_size
        )
    
    @trace_function(name="vllm_generate")
    def generate(
        self,
        prompt: str,
        config: Optional[GenerationConfig] = None
    ) -> GenerationResult:
        """Generate text using vLLM."""
        results = self.generate_batch([prompt], config)
        return results[0]
    
    def generate_batch(
        self,
        prompts: List[str],
        config: Optional[GenerationConfig] = None
    ) -> List[GenerationResult]:
        """Generate text for multiple prompts using vLLM."""
        self._load_model()
        
        try:
            from vllm import SamplingParams
        except ImportError:
            raise ImportError("vllm is required")
        
        config = config or GenerationConfig()
        
        sampling_params = SamplingParams(
            max_tokens=config.max_tokens,
            temperature=config.temperature,
            top_p=config.top_p,
            top_k=config.top_k,
            repetition_penalty=config.repetition_penalty
        )
        
        outputs = self._llm.generate(prompts, sampling_params)
        
        results = []
        for output in outputs:
            text = output.outputs[0].text
            results.append(GenerationResult(
                text=text.strip(),
                tokens_used=len(output.outputs[0].token_ids),
                model=self.model_name
            ))
        
        return results


def create_llm_client(settings) -> LLMClient:
    """
    Factory function to create LLM client from settings.
    
    Args:
        settings: LLMSettings or Settings object
        
    Returns:
        LLMClient instance
    """
    if hasattr(settings, "llm"):
        llm_settings = settings.llm
    else:
        llm_settings = settings
    
    backend = llm_settings.backend
    
    if backend == "huggingface":
        return HuggingFaceLLM(
            model_name=llm_settings.model_name,
            device=llm_settings.device,
            quantization=llm_settings.quantization,
            trust_remote_code=llm_settings.huggingface.trust_remote_code,
            torch_dtype=llm_settings.huggingface.torch_dtype
        )
    elif backend == "ollama":
        return OllamaLLM(
            model_name=llm_settings.model_name,
            host=llm_settings.ollama.host
        )
    elif backend == "openai":
        return OpenAICompatibleLLM(
            model_name=llm_settings.model_name,
            api_base=llm_settings.openai.api_base,
            api_key=llm_settings.openai.api_key
        )
    elif backend == "vllm":
        return VLLMClient(
            model_name=llm_settings.model_name,
            tensor_parallel_size=llm_settings.vllm.tensor_parallel_size
        )
    else:
        raise ValueError(f"Unknown LLM backend: {backend}")



"""Generation modules for cheat sheet content synthesis."""

from .concepts import ConceptExtractor
from .formulas import FormulaExtractor
from .models_theory import ModelSummarizer
from .algorithms import AlgorithmSynthesizer
from .latex_builder import LaTeXBuilder, CheatSheet

__all__ = [
    "ConceptExtractor",
    "FormulaExtractor",
    "ModelSummarizer",
    "AlgorithmSynthesizer",
    "LaTeXBuilder",
    "CheatSheet",
]
